"""
This module implements a mock of the kubernetes client library which can be used
to patch the api client in an ansible module.

We attempt to emulate the internals of the kubernetes api_client, but this is
based on code inspection of the current implementation and is certainly subject
to change!
"""

# Standard
from contextlib import contextmanager
from datetime import datetime, timedelta
from queue import Empty, Queue
from threading import Event, RLock
from typing import Optional
from unittest import mock
import base64
import json

# Third Party
from openshift.dynamic.apply import annotate as annotate_last_applied
import kubernetes

# First Party
import aconfig
import alog

# Local
from oper8.deploy_manager import KubeEventType

log = alog.use_channel("TEST")


class MockKubRestResponse(kubernetes.client.rest.RESTResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def getheaders():
        return {}


class MockWatchStreamResponse:
    """Helper class used to stream resources from the MockKubeClient using a queue.
    When the streaming method is called the MockWatchStreamResponse registers a
    threading queue with the MockKubClient and yields all events"""

    def __init__(
        self,
        api_client: "MockKubClient",
        api_version: str,
        kind: str,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        timeout: Optional[int] = None,
    ):
        self.api_client = api_client
        self.watch_queue = Queue()
        self.timeout = timeout or 250

        self.kind = kind
        self.api_version = api_version
        self.namespace = namespace
        self.name = name

        # Shutdown flag
        self.shutdown = Event()

    def __del__(self):
        self.api_client._remove_watch_queue(self.watch_queue)

    def stream(self, *args, **kwargs):
        """Continuously yield events from the cluster until the shutdown or timeout"""

        # Get the current resource state
        current_resources = []
        if self.name:
            current_obj = self.api_client._get_object_state(
                method="GET",
                namespace=self.namespace,
                kind=self.kind,
                api_version=self.api_version,
                name=self.name,
            )
            if current_obj:
                current_resources.append(current_obj)
        else:
            response, code = self.api_client._list_object_state(
                method="GET",
                namespace=self.namespace,
                kind=self.kind,
                api_version=self.api_version,
            )
            current_resources = response.get("items")

        # yield back the resources
        for resource in current_resources:
            log.debug2("Yielding initial state event")
            yield self._make_watch_response(KubeEventType.ADDED, resource)

        log.debug2("Yielded initial state. Starting watch")
        # Create a watch queue and add it to the api_client
        self.api_client._add_watch_queue(self.watch_queue)

        # Configure the timeout and end times
        timeout_delta = timedelta(seconds=self.timeout)
        end_time = datetime.now() + timeout_delta
        while True:
            timeout = (end_time - datetime.now()).total_seconds() or 1
            try:
                event_type, resource = self.watch_queue.get(timeout=timeout)
            except Empty:
                return

            if self._check_end_conditions(end_time):
                return

            resource_metadata = resource.get("metadata", {})

            # Ensure the kind/apiversion/namespace match the requested
            if (
                resource.get("kind") == self.kind
                and resource.get("apiVersion") == self.api_version
                and resource_metadata.get("namespace") == self.namespace
            ):
                # If resourced then ensure the name matches
                if self.name and resource_metadata.get("name") != self.name:
                    continue

                log.debug2("Yielding watch event")
                yield self._make_watch_response(event_type, resource)

    def close(self):
        pass

    def release_conn(self):
        self.shutdown.set()

    def _check_end_conditions(self, end_time):
        log.debug3("Checking shutdown and endtime conditions")
        if self.shutdown.is_set():
            return True

        return end_time < datetime.now()

    def _make_watch_response(self, event, object):
        # Add new line to watch response
        response = json.dumps({"type": event.value, "object": object}) + "\n"
        return response


class MockKubClient(kubernetes.client.ApiClient):
    """Mocked version of kubernetes.client.ApiClient which swaps out the
    implementation of call_api() to use preconfigured responses
    """

    def __init__(self, cluster_state=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Save the current cluster state
        self.__cluster_state = cluster_state or {}
        log.debug3("Cluster State: %s", self.__cluster_state)

        # Setup listener variables
        self.__queue_lock = RLock()
        self.__watch_queues = set()

        # Save a list of kinds for each api group
        self._api_group_kinds = {}

        # Canned handlers
        self._handlers = {
            "/apis": {"GET": self.apis},
            "/version": {"GET": lambda *_, **__: self._make_response({})},
            "/api/v1": {"GET": self.api_v1},
        }
        for namespace, ns_entries in self.__cluster_state.items():
            for kind, kind_entries in ns_entries.items():
                for api_version, version_entries in kind_entries.items():
                    # Add handlers common to this kind
                    self._add_handlers_for_kind(namespace, kind, api_version)

                    # Add handlers for the individual pre-existing instances
                    for name, obj_state in version_entries.items():
                        # If this is a static object state, make sure the
                        # metadata aligns correctly
                        if isinstance(obj_state, dict):
                            self._add_resource_defaults(
                                obj_state=obj_state,
                                namespace=namespace,
                                kind=kind,
                                api_version=api_version,
                                name=name,
                            )

                        # Add the endpoints for this resource
                        self._add_handlers_for_resource(
                            namespace=namespace,
                            kind=kind,
                            api_version=api_version,
                            name=name,
                        )

        log.debug("Configured Handlers: %s", list(self._handlers.keys()))

    def call_api(
        self,
        resource_path,
        method,
        path_params=None,
        query_params=None,
        header_params=None,
        body=None,
        **kwargs,
    ):
        """Mocked out call function to return preconfigured responses

        NOTE: this is set up to work with how openshift.dynamic.DynamicClient
            calls self.client.call_api. It (currently) passes some as positional
            args and some as kwargs.
        """
        for key, value in query_params:
            if key == "watch" and value:
                method = "WATCH"
                break

        log.debug2("Mock [%s] request to [%s]", method, resource_path)
        log.debug4("Path Params: %s", path_params)
        log.debug4("Query Params: %s", query_params)
        log.debug4("Header Params: %s", header_params)
        log.debug4("Body: %s", body)

        # Find the right handler and execute it
        return self._get_handler(resource_path, method)(
            resource_path=resource_path,
            path_params=path_params,
            query_params=query_params,
            header_params=header_params,
            body=body,
            **kwargs,
        )

    ## Implementation Helpers ##################################################

    @staticmethod
    def _make_response(body, status_code=200):
        log.debug2("Making response with code: %d", status_code)
        log.debug4(body)
        resp = MockKubRestResponse(
            aconfig.Config(
                {
                    "status": status_code,
                    "reason": "MOCK",
                    "data": json.dumps(body).encode("utf8"),
                }
            )
        )
        if not 200 <= status_code <= 299:
            raise kubernetes.client.rest.ApiException(http_resp=resp)
        return resp

    def _get_handler(self, resource_path, method):
        # Look for a configured handler that matches the path exactly
        handler = self._handlers.get(resource_path, {}).get(method)

        # If no handler found, start looking for '*' handlers
        path_parts = resource_path.split("/")
        while handler is None and path_parts:
            star_path = "/".join(path_parts + ["*"])
            log.debug4("Looking for [%s]", star_path)
            handler = self._handlers.get(star_path, {}).get(method)
            path_parts.pop()

        # Return whatever we've found or not_found
        return handler or self.not_found

    @staticmethod
    def _get_kind_variants(kind):
        return [kind.lower(), f"{kind.lower()}s", kind]

    @staticmethod
    def _get_version_parts(api_version):
        parts = api_version.split("/", 1)
        if len(parts) == 2:
            group_name, version = parts
            api_endpoint = f"/apis/{group_name}/{version}"
        else:
            group_name = None
            version = api_version
            api_endpoint = f"/api/{api_version}"
        return group_name, version, api_endpoint

    def _add_handlers_for_kind(self, namespace, kind, api_version):
        # Set up the information needed for the apis and crds calls
        group_name, version, api_endpoint = self._get_version_parts(api_version)
        self._api_group_kinds.setdefault(group_name, {}).setdefault(version, []).append(
            kind
        )

        # Add a configured handler for this group type
        log.debug("Adding resource handler for: %s", api_endpoint)
        self._handlers[api_endpoint] = {
            "GET": lambda *_, **__: (self.current_state_crds(group_name, version))
        }

        # Add POST handlers for this type
        for kind_variant in self._get_kind_variants(kind):
            # Add different endpoints based on namespace
            if namespace:
                endpoint = f"{api_endpoint}/namespaces/{namespace}/{kind_variant}/*"
            else:
                endpoint = f"{api_endpoint}/{kind_variant}/*"

            log.debug2(
                "Adding POST & PUT & GET & WATCH & PATCH handler for (%s: %s)",
                kind,
                endpoint,
            )
            self._handlers[endpoint] = {
                "WATCH": lambda resource_path, body, *_, **__: (
                    self.current_state_watch(
                        resource_path, api_version, kind, resourced=False
                    )
                ),
                "PATCH": lambda resource_path, body, *_, **__: (
                    self.current_state_patch(resource_path, api_version, kind, body)
                ),
                "PUT": lambda resource_path, body, *_, **__: (
                    self.current_state_patch(resource_path, api_version, kind, body)
                ),
                "POST": lambda resource_path, body, *_, **__: (
                    self.current_state_post(resource_path, api_version, kind, body)
                ),
                "GET": lambda resource_path, *_, **__: (
                    self.current_state_list(resource_path, api_version, kind)
                ),
            }

    def _remove_handlers_for_resource(self, namespace, kind, api_version, name):
        # Get crucial API information out of the object
        _, __, api_endpoint = self._get_version_parts(api_version)
        for kind_variant in self._get_kind_variants(kind):
            if namespace:
                resource_api_endpoint = (
                    f"{api_endpoint}/namespaces/{namespace}/{kind_variant}/{name}"
                )
            else:
                resource_api_endpoint = f"{api_endpoint}/{kind_variant}/{name}"

            status_resource_api_endpoint = "/".join([resource_api_endpoint, "status"])
            del self._handlers[resource_api_endpoint]
            del self._handlers[status_resource_api_endpoint]

    def _add_handlers_for_resource(self, namespace, kind, api_version, name):
        # Get crucial API information out of the object
        _, __, api_endpoint = self._get_version_parts(api_version)

        # Add configured handlers for GET/PUT on this resource
        for kind_variant in self._get_kind_variants(kind):
            # The endpoint that will be used to hit this specific
            # resource
            if namespace:
                resource_api_endpoint = (
                    f"{api_endpoint}/namespaces/{namespace}/{kind_variant}/{name}"
                )
            else:
                resource_api_endpoint = f"{api_endpoint}/{kind_variant}/{name}"

            # Add the handlers
            log.debug("Adding GET handler for: %s", resource_api_endpoint)
            self._handlers[resource_api_endpoint] = {
                "GET": lambda *_, x=resource_api_endpoint, **__: (
                    self.current_state_get(x, api_version, kind)
                ),
                "PUT": lambda body, *_, x=resource_api_endpoint, **__: (
                    self.current_state_put(x, api_version, kind, body)
                ),
                "PATCH": lambda body, *_, x=resource_api_endpoint, **__: (
                    self.current_state_patch(x, api_version, kind, body)
                ),
                "DELETE": lambda *_, x=resource_api_endpoint, **__: (
                    self.current_state_delete(x, api_version, kind)
                ),
                "WATCH": lambda resource_path, body, *_, **__: (
                    self.current_state_watch(
                        resource_path, api_version, kind, resourced=True
                    )
                ),
            }

            # Add status PUT
            status_resource_api_endpoint = "/".join([resource_api_endpoint, "status"])
            self._handlers[status_resource_api_endpoint] = {
                "PUT": lambda body, *_, x=status_resource_api_endpoint, **__: (
                    self.current_state_put(
                        x,
                        api_version,
                        kind,
                        body,
                        is_status=True,
                    )
                )
            }

    def _get_object_state(self, method, namespace, kind, api_version, name):
        create = method in ["PUT", "PATCH", "POST"]
        if create:
            content = (
                self.__cluster_state.setdefault(namespace, {})
                .setdefault(kind, {})
                .setdefault(api_version, {})
                .get(name, {})
            )
        else:
            content = (
                self.__cluster_state.get(namespace, {})
                .get(kind, {})
                .get(api_version, {})
                .get(name)
            )

        # If it's a callable, call it!
        if callable(content):
            log.debug2("Making callable resource content")
            content = content(
                method=method,
                namespace=namespace,
                kind=kind,
                api_version=api_version,
                name=name,
            )

            # Add the defaults and handle the case where it's a tuple with a
            # status code
            if isinstance(content, tuple):
                body, status = content
                log.debug3("Handling tuple content with status [%s]", status)
                self._add_resource_defaults(body, namespace, kind, api_version, name)
                content = (body, status)
            else:
                self._add_resource_defaults(content, namespace, kind, api_version, name)
            log.debug3("Content: %s", content)
        return content

    def _list_object_state(self, method, namespace, kind, api_version):
        if method != "GET":
            return ([], 405)

        content = (
            self.__cluster_state.setdefault(namespace, {})
            .setdefault(kind, {})
            .setdefault(api_version, {})
        )

        resource_list = []
        return_status = 200
        for resource_name in content:
            resource = content[resource_name]

            # If it's a callable, call it!
            if callable(resource):
                log.debug2("Making callable resource content")
                resource = resource(
                    method=method,
                    namespace=namespace,
                    kind=kind,
                    api_version=api_version,
                    name=resource_name,
                )

            # Add the defaults and handle the case where it's a tuple with a
            # status code
            if isinstance(resource, tuple):
                resource, status = resource
                if status == 403:
                    return_status = 403
                    break
                self._add_resource_defaults(
                    resource, namespace, kind, api_version, resource_name
                )
            else:
                self._add_resource_defaults(
                    resource, namespace, kind, api_version, resource_name
                )

            log.debug3("Resource: %s", content)
            resource_list.append(resource)

        content = {"apiVersion": "v1", "kind": "List", "items": resource_list}
        return (content, return_status)

    def _update_object_current_state(self, namespace, kind, api_version, name, state):
        """Helper function to update a resource in the cluster and update all watch queues"""
        self.__cluster_state.setdefault(namespace, {}).setdefault(kind, {}).setdefault(
            api_version, {}
        )

        # Get the event type based on if name already exists
        event_type = KubeEventType.ADDED
        if name in self.__cluster_state[namespace][kind][api_version]:
            event_type = KubeEventType.MODIFIED

        # Update the cluster state and add resource_path if it doesn't already exist
        self.__cluster_state[namespace][kind][api_version][name] = state
        self._add_handlers_for_resource(
            namespace=namespace, kind=kind, api_version=api_version, name=name
        )
        self._update_watch_queues(event_type, state)

    def _delete_object_state(self, namespace, kind, api_version, name):
        """Helper function to delete a resource in the cluster and update all watch queues"""

        original_object = self.__cluster_state[namespace][kind][api_version][name]

        del self.__cluster_state[namespace][kind][api_version][name]
        if not self.__cluster_state[namespace][kind][api_version]:
            del self.__cluster_state[namespace][kind][api_version]
        if not self.__cluster_state[namespace][kind]:
            del self.__cluster_state[namespace][kind]
        if not self.__cluster_state[namespace]:
            del self.__cluster_state[namespace]

        self._update_watch_queues(KubeEventType.DELETED, original_object)

        # Remove any endpoint handlers for this resource
        self._remove_handlers_for_resource(namespace, kind, api_version, name)

        return True

    def _add_watch_queue(self, queue):
        with self.__queue_lock:
            log.debug3("Adding watch queue %s", queue)
            self.__watch_queues.add(queue)

    def _remove_watch_queue(self, queue):
        with self.__queue_lock:
            log.debug3("Removing watch queue %s", queue)
            self.__watch_queues.remove(queue)

    def _update_watch_queues(self, event, object):
        with self.__queue_lock:
            log.debug2("Updating watch queues with %s event", event)
            for queue in self.__watch_queues:
                queue.put((event, object))

    @staticmethod
    def _add_resource_defaults(obj_state, namespace, kind, api_version, name):
        obj_state["apiVersion"] = api_version
        obj_state["kind"] = kind
        md = obj_state.setdefault("metadata", {})
        if namespace:
            md["namespace"] = namespace
        md["name"] = name
        last_applied_annotation = annotate_last_applied(obj_state)
        md.setdefault("annotations", {}).update(
            last_applied_annotation["metadata"]["annotations"]
        )

    @classmethod
    def _patch_resource(cls, base, overrides):
        """Merge helper that supports removing elements when the override is set
        to None
        """
        for key, value in overrides.items():
            if value is None and key in base:
                del base[key]

            elif (
                key not in base
                or not isinstance(base[key], dict)
                or not isinstance(value, dict)
            ):
                base[key] = value
            else:
                base[key] = cls._patch_resource(base[key], value)

        return base

    ## Handlers ################################################################

    @classmethod
    def not_found(cls, *_, **__):
        log.debug3("Not Found")
        return cls._make_response(
            {
                "kind": "Status",
                "apiVersion": "v1",
                "metadata": {},
                "status": "Failure",
                "message": "the server could not find the requested resource",
                "reason": "NotFound",
                "details": {},
                "code": 404,
            },
            404,
        )

    def apis(self, *_, **__):
        api_group_list = {"kind": "APIGroupList", "apiVersion": "v1", "groups": []}
        for group_name, api_versions in self._api_group_kinds.items():
            if group_name is None:
                continue
            group = {"name": group_name, "versions": []}
            for api_version in api_versions:
                group["versions"].append(
                    {
                        "groupVersion": f"{group_name}/{api_version}",
                        "version": api_version,
                    }
                )
            group["preferredVersion"] = group["versions"][0]
            api_group_list["groups"].append(group)
        return self._make_response(api_group_list)

    def api_v1(self, *_, **__):
        return self.current_state_crds(None, "v1")

    def current_state_crds(self, group_name, api_version):
        resource_list = {
            "kind": "APIResourceList",
            "apiVersion": "v1",
            "groupVersion": f"{group_name}/{api_version}",
            "resources": [],
        }
        for kind in self._api_group_kinds.get(group_name, {}).get(api_version, []):
            resource_list["resources"].append(
                {
                    "name": f"{kind.lower()}s",
                    "singularName": kind.lower(),
                    "namespaced": True,
                    "kind": kind,
                    "verbs": [
                        "delete",
                        "deletecollection",
                        "get",
                        "list",
                        "patch",
                        "create",
                        "update",
                        "watch",
                    ],
                    "storageVersionHash": base64.b64encode(kind.encode("utf-8")).decode(
                        "utf-8"
                    ),
                }
            )
            resource_list["resources"].append(
                {
                    "name": f"{kind.lower()}s/status",
                    "singularName": "",
                    "namespaced": True,
                    "kind": kind,
                    "verbs": ["get", "patch", "update"],
                }
            )
        return self._make_response(resource_list)

    def current_state_watch(
        self, api_endpoint, api_version, kind, resourced=False, query_params=None
    ):
        # Parse the endpoint for the namespace and name
        endpoint_parts = api_endpoint.split("/")
        namespace = None
        name = None
        if "namespaces" in endpoint_parts:
            namespace = endpoint_parts[endpoint_parts.index("namespaces") + 1]

        if resourced:
            name = endpoint_parts[-1]

        # Return Watch Stream Response
        return MockWatchStreamResponse(
            api_client=self,
            api_version=api_version,
            kind=kind,
            namespace=namespace,
            name=name,
            timeout=(query_params or {}).get("timeoutSeconds"),
        )

    def current_state_get(self, api_endpoint, api_version, kind):
        # Parse the endpoint for the namespace and name
        endpoint_parts = api_endpoint.split("/")
        namespace = ""
        if "namespaces" in endpoint_parts:
            namespace = endpoint_parts[endpoint_parts.index("namespaces") + 1]
        name = endpoint_parts[-1]

        # Look up the resources in the cluster state
        log.debug2(
            "Looking for current state of [%s/%s/%s/%s]",
            namespace,
            kind,
            api_version,
            name,
        )
        content = self._get_object_state(
            method="GET",
            namespace=namespace,
            kind=kind,
            api_version=api_version,
            name=name,
        )
        log.debug4("Content: %s", content)
        if content is not None:
            # If the content includes a status code, make the response with it
            if isinstance(content, tuple):
                return self._make_response(*content)
            return self._make_response(content)
        return self.not_found()

    def current_state_list(self, api_endpoint, api_version, kind):
        # Parse the endpoint for the namespace and name and where the kind is located
        # in endpoint_parts
        endpoint_parts = api_endpoint.split("/")
        namespace = ""
        kind_loc = 1
        if "namespaces" in endpoint_parts:
            kind_loc = endpoint_parts.index("namespaces") + 2
            namespace = endpoint_parts[kind_loc - 1]
        else:
            version_split = api_version.split("/")
            # 2 for ["","api"] and then add length of api_version split which would be
            # 2 for resources with a group and 1 without e.g. v1 = 1 and foo.bar.com/v1 would be 2
            kind_loc = 2 + len(version_split)

        # If Api was trying to get a specific resource and not list then return 404
        # as object must not have been found. This is checked by seeing if the kind
        # is at the end of the endpoint_parts
        if kind_loc != len(endpoint_parts) - 1:
            return self.not_found()

        # Look up the resources in the cluster state
        log.debug2(
            "Listing current state of [%s/%s/%s]",
            namespace,
            kind,
            api_version,
        )
        content = self._list_object_state(
            method="GET",
            namespace=namespace,
            kind=kind,
            api_version=api_version,
        )
        log.debug4("Content: %s", content)
        if content is not None:
            # If the content includes a status code, make the response with it
            if isinstance(content, tuple):
                return self._make_response(*content)
            return self._make_response(content)
        return self.not_found()

    def current_state_put(self, api_endpoint, api_version, kind, body, is_status=False):
        # Parse the endpoint for the namespace and name
        endpoint_parts = api_endpoint.split("/")
        namespace = ""
        if "namespaces" in endpoint_parts:
            namespace = endpoint_parts[endpoint_parts.index("namespaces") + 1]
        name = endpoint_parts[-1] if not is_status else endpoint_parts[-2]

        # Look up the resources in the cluster state
        log.debug2(
            "Looking for current state of [%s/%s/%s/%s]",
            namespace,
            kind,
            api_version,
            name,
        )
        content = self._get_object_state(
            method="PUT",
            namespace=namespace,
            kind=kind,
            api_version=api_version,
            name=name,
        )
        log.debug3("Current Content: %s", content)

        # If the content has a status code, unpack it
        status_code = 200
        if isinstance(content, tuple):
            content, status_code = content

        # If it's a non-200 status code, don't make the update
        if status_code != 200:
            return self._make_response(content, status_code)

        # If this is a status, we are only updating the status and keeping the
        # existing content
        if is_status:
            content.update({"status": body.get("status", {})})
            updated_content = content
        else:
            if "status" in body:
                del body["status"]
            updated_content = body
        log.debug3(
            "Updating [%s/%s/%s/%s] with body: %s",
            namespace,
            kind,
            api_version,
            name,
            updated_content,
        )
        self._update_object_current_state(
            namespace, kind, api_version, name, updated_content
        )

        return self._make_response(updated_content, status_code)

    def current_state_patch(self, api_endpoint, api_version, kind, body):
        # Parse the endpoint for the namespace and name
        endpoint_parts = api_endpoint.split("/")
        namespace = ""
        if "namespaces" in endpoint_parts:
            namespace = endpoint_parts[endpoint_parts.index("namespaces") + 1]
        name = endpoint_parts[-1]

        # Look up the resources in the cluster state
        log.debug2(
            "Looking for current state of [%s/%s/%s/%s]",
            namespace,
            kind,
            api_version,
            name,
        )
        content = self._get_object_state(
            method="PATCH",
            namespace=namespace,
            kind=kind,
            api_version=api_version,
            name=name,
        )
        log.debug3("Current Content: %s", content)
        log.debug3("Update body: %s", body)

        # If the content has a status code, unpack it
        status_code = 200
        if isinstance(content, tuple):
            content, status_code = content

        # If it's a non-200 status code, don't make the update
        if status_code != 200:
            return self._make_response(content, status_code)

        # Merge in the new body
        if "status" in body:
            del body["status"]
        log.debug3(
            "Updating [%s/%s/%s/%s] with body: %s",
            namespace,
            kind,
            api_version,
            name,
            body,
        )
        updated_content = self._patch_resource(content, body)
        log.debug3("Updated content: %s", updated_content)
        self._update_object_current_state(
            namespace, kind, api_version, name, updated_content
        )

        return self._make_response(updated_content, status_code)

    def current_state_post(self, api_endpoint, api_version, kind, body):
        log.debug2("Creating current state for [%s]", api_endpoint)

        # Parse the endpoint and body for the namespace and name
        endpoint_parts = api_endpoint.split("/")
        namespace = ""
        if "namespaces" in endpoint_parts:
            namespace = endpoint_parts[endpoint_parts.index("namespaces") + 1]
        name = body.get("metadata", {}).get("name")

        # Look up the resources in the cluster state
        log.debug2(
            "Looking for current state of [%s/%s/%s/%s]",
            namespace,
            kind,
            api_version,
            name,
        )
        content = self._get_object_state(
            method="POST",
            namespace=namespace,
            kind=kind,
            api_version=api_version,
            name=name,
        )
        log.debug3("Current Content: %s", content)

        # If the content has a status code, unpack it
        status_code = 200
        if isinstance(content, tuple):
            content, status_code = content

        # If it's a non-200 status code, don't make the update
        if status_code != 200:
            return self._make_response(content, status_code)

        # Overwrite the body
        log.debug3("Overwrite content: %s", body)
        self._update_object_current_state(namespace, kind, api_version, name, body)

        return self._make_response(body, status_code)

    def current_state_delete(self, api_endpoint, api_version, kind):
        # Parse the endpoint for the namespace and name
        endpoint_parts = api_endpoint.split("/")
        namespace = ""
        if "namespaces" in endpoint_parts:
            namespace = endpoint_parts[endpoint_parts.index("namespaces") + 1]
        name = endpoint_parts[-1]

        # Look up the resources in the cluster state
        log.debug2(
            "Looking for current state of [%s/%s/%s/%s]",
            namespace,
            kind,
            api_version,
            name,
        )

        content = self._get_object_state("DELETE", namespace, kind, api_version, name)
        response_code = 200
        if isinstance(content, tuple):
            content, response_code = content
        deleted = content is not None

        if content is not None:
            self._delete_object_state(
                namespace=namespace,
                kind=kind,
                api_version=api_version,
                name=name,
            )
        response_content = {
            "kind": "status",
            "apiVersion": "v1",
            "metadata": {},
            "details": {
                "name": name,
                "kind": kind,
            },
        }

        if deleted:
            response_content["status"] = "Success"
            response_content["details"]["uid"] = "Hope nothing uses this"
        else:
            response_content["status"] = "Failure"
            response_content["message"] = f'{kind} "{name}" not found'
            response_code = 404
        return self._make_response(response_content, response_code)


@contextmanager
def mock_kub_client_constructor(*args, **kwargs):
    """Context manager to patch the api client"""
    log.debug("Getting mocked client")
    client = MockKubClient(*args, **kwargs)
    log.debug("Mock client complete")
    with mock.patch(
        "kubernetes.config.new_client_from_config",
        return_value=client,
    ):
        yield client



"""
Tests for the ansible watch manager.

NOTE: Without access to the actual ansible-operator executable, these tests are
    primarily intended to ensure the logic in the class works. To validate that
    the watches actually work, we need to build and run an operator!
"""

# Standard
import json
import os
import shutil
import sys
import tempfile
import time

# Third Party
import dill
import pytest
import yaml

# First Party
import alog

# Local
from oper8.test_helpers.helpers import DummyController, library_config
from oper8.watch_manager import WatchManagerBase
from oper8.watch_manager.ansible_watch_manager import AnsibleWatchManager

## Helpers #####################################################################

log = alog.use_channel("TEST")


def get_entrypoint(func):
    """Get an entrypoint command that will run this file as a main and pass it
    the pickled version of the given function
    """
    this_file = os.path.realpath(__file__)
    func_str = dill.dumps(func).hex()
    return f'{sys.executable} {this_file} "{func_str}"'


# This main is used to simulate the ansible operator executable and run various
# tests against it
if __name__ == "__main__":
    # Grab the first argument as a pickled string and deserialize it into a
    # function
    func = dill.loads(bytes.fromhex(sys.argv[1]))

    # Run the function, letting failures crash as needed
    func(*sys.argv[2:])


class FinalizerController(DummyController):
    kind = "Finalizable"

    def finalize_components(self, session):
        pass


def reset_globals(workdir=None):
    """This helper is only used in tests to "reset" the state of the global
    ansible process so that tests can run independently
    """
    AnsibleWatchManager.ANSIBLE_PROCESS = None
    WatchManagerBase._ALL_WATCHES = {}
    if workdir is not None:
        shutil.rmtree(workdir)
        os.mkdir(workdir)


@pytest.fixture(autouse=True)
def reset_globals_fxtr():
    """This helper is only used in tests to "reset" the state of the global
    ansible process so that tests can run independently
    """
    AnsibleWatchManager.ANSIBLE_PROCESS = None
    WatchManagerBase._ALL_WATCHES = {}


@pytest.fixture
def workdir():
    """All the tests need a temporary workdir, so we do this as a fixture"""
    with tempfile.TemporaryDirectory() as workdir:
        yield workdir


## Tests #######################################################################

#################
## Constructor ##
#################


@pytest.mark.ansible
def test_constructor_playbook(workdir):
    """Test that the playbook gets created with all passed-in parameters"""
    playbook_params = {
        "foo": True,
        "bar": "baz",
    }
    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        playbook_parameters=playbook_params,
    )

    # Make sure the playbook exists
    playbook_path = os.path.join(workdir, "playbook-foo.yaml")
    assert os.path.exists(playbook_path)

    # Read the playbook
    with open(playbook_path, "r") as handle:
        parsed_playbook = yaml.safe_load(handle)
    log.debug3(parsed_playbook)

    # Make sure passthrough args are passed through
    vars = parsed_playbook[0]["tasks"][0]["vars"]
    for key, val in playbook_params.items():
        assert vars.get(key) == val

    # Check the necessary additions made to wire up the controller
    assert vars["full_cr"] == "{{ _foo_bar_com_foo }}"
    assert "controller_class" in vars
    assert "strict_versioning" in vars


@pytest.mark.ansible
def test_constructor_watch_entry(workdir):
    """Test that a watch entry is added by the constructor"""
    playbook_params = {
        "foo": True,
        "bar": "baz",
    }
    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        manage_status=True,
        watch_dependent_resources=True,
        reconcile_period="3m",
        playbook_parameters=playbook_params,
    )

    # Make sure the watches file exists
    watches_path = os.path.join(workdir, "watches.yaml")
    assert os.path.exists(watches_path)

    # Read the watches file
    with open(watches_path, "r") as handle:
        parsed_watches = yaml.safe_load(handle)
    log.debug3(parsed_watches)

    # Make sure all elements are set correctly
    assert len(parsed_watches) == 1
    watch = parsed_watches[0]
    assert watch.get("group") == wm.group
    assert watch.get("version") == wm.version
    assert watch.get("kind") == wm.kind
    assert watch.get("playbook") == os.path.join(workdir, "playbook-foo.yaml")
    assert watch.get("manageStatus")
    assert watch.get("watchDependentResources")
    assert watch.get("reconcilePeriod") == "3m"
    assert watch.get("vars", {}).get("operation") == "add"
    assert "finalizer" not in watch


@pytest.mark.ansible
def test_constructor_watch_entry_finalizer(workdir):
    """Test that a watch entry for a controller with a finalier has the needed
    watches.yaml entries.
    """
    wm = AnsibleWatchManager(FinalizerController, ansible_base_path=workdir)

    # Make sure the watches file exists
    watches_path = os.path.join(workdir, "watches.yaml")
    assert os.path.exists(watches_path)

    # Read the watches file
    with open(watches_path, "r") as handle:
        parsed_watches = yaml.safe_load(handle)
    log.debug3(parsed_watches)

    # Make sure all elements are set correctly
    assert len(parsed_watches) == 1
    watch = parsed_watches[0]
    assert watch.get("vars", {}).get("operation") == "add"
    assert watch.get("finalizer", {}).get("name") == (
        f"finalizers.{FinalizerController.kind.lower()}.{FinalizerController.group}"
    )
    assert watch.get("finalizer", {}).get("vars", {}).get("operation") == "remove"


@pytest.mark.ansible
def test_constructor_multiple_watches(workdir):
    """Test that multiple watches can be added to the same watches.yaml by
    different managers
    """

    class DummyController2(DummyController):
        group = "foo.bar"
        version = "v2"
        kind = "Bar"

    # Create two watch managers
    wm1 = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
    )
    wm2 = AnsibleWatchManager(
        DummyController2,
        ansible_base_path=workdir,
    )

    # Make sure the watches file exists
    watches_path = os.path.join(workdir, "watches.yaml")
    assert os.path.exists(watches_path)

    # Read the watches file
    with open(watches_path, "r") as handle:
        parsed_watches = yaml.safe_load(handle)
    log.debug3(parsed_watches)

    # Make sure there are two entries in the watches file
    assert len(parsed_watches) == 2


@pytest.mark.ansible
def test_constructor_no_duplicate_watches(workdir):
    """Test that an assertion is raised if two managers attempt to watch the
    same group/version/kind
    """
    # Create two watch managers
    wm1 = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
    )
    with pytest.raises(AssertionError):
        AnsibleWatchManager(
            DummyController,
            ansible_base_path=workdir,
        )


@pytest.mark.ansible
def test_constructor_watch_disable_vcs(workdir):
    """Test that a watched controller with disable_vcs set will add the
    correct variable to the watch.yaml section
    """

    class NonVersionedController(DummyController):
        group = "foo.bar"
        version = "v2"
        kind = "Bar"
        disable_vcs = True

    wm = AnsibleWatchManager(
        NonVersionedController,
        ansible_base_path=workdir,
    )

    # Make sure the watches file exists
    watches_path = os.path.join(workdir, "watches.yaml")
    assert os.path.exists(watches_path)

    # Read the watches file
    with open(watches_path, "r") as handle:
        parsed_watches = yaml.safe_load(handle)
    log.debug3(parsed_watches)

    # Make sure the enable_ansible_vcs variable is set correctly
    assert len(parsed_watches) == 1
    assert parsed_watches[0]["vars"]["enable_ansible_vcs"] == "false"


@pytest.mark.ansible
def test_constructor_log_dir(workdir):
    """Test that if a log_dir is given in config, it is used in the playbook"""
    log_dir = "/logs"
    with library_config(ansible_watch_manager={"log_file_dir": log_dir}):
        wm = AnsibleWatchManager(
            DummyController,
            ansible_base_path=workdir,
        )

        # Make sure the playbook exists
        playbook_path = os.path.join(workdir, "playbook-foo.yaml")
        assert os.path.exists(playbook_path)

        # Read the playbook
        with open(playbook_path, "r") as handle:
            parsed_playbook = yaml.safe_load(handle)
        log.debug3(parsed_playbook)

        # Make sure log_file starts with the log dir
        vars = parsed_playbook[0]["tasks"][0]["vars"]
        assert vars.get("log_file", "").startswith(log_dir)


@pytest.mark.ansible
def test_constructor_arg_override_precedence(workdir):
    """Test that all of the kwargs in the initializer adhere to the correct
    override precedence
    """
    # Set up config overrides with the correct types and with all strings to
    # simulate reading from the env
    config_overrides = {
        "ansible_base_path": workdir,
        "ansible_entrypoint": "foobar.sh",
        "ansible_args": "--ansible-args string",
        "manage_status": True,
        "watch_dependent_resources": True,
        "reconcile_period": "120m",
        "playbook_parameters": {"foo": "bar"},
    }
    env_config_overrides = {
        key: (str(val) if not isinstance(val, dict) else json.dumps(val))
        for key, val in config_overrides.items()
    }

    # Set up overrides that will be ignored by overrides given in kwargs
    ignored_config_overrides = {
        "ansible_base_path": "SOMETHING BAD",
        "ansible_entrypoint": "yikes",
        "manage_status": "this can't even convert to a bool",
        "watch_dependent_resources": "neither can this!",
        "reconcile_period": "120m!!!!!!!!!!!!",
        "playbook_parameters": "no dict here",
    }

    # Test that config overrides take precedence over defaults
    for override_set in [
        config_overrides,
        env_config_overrides,
        ignored_config_overrides,
    ]:
        with library_config(ansible_watch_manager=override_set):
            # Reset the globals for this override set
            reset_globals(workdir)

            # Make the watch manager using kwargs only if overrides not used
            # from config
            kwargs = {}
            if override_set is ignored_config_overrides:
                kwargs = config_overrides
            wm = AnsibleWatchManager(DummyController, **kwargs)

            # Make sure the entrypoint and base path were set correctly
            assert wm._ansible_base_path == config_overrides["ansible_base_path"]
            assert wm._ansible_entrypoint == config_overrides["ansible_entrypoint"]
            assert wm._ansible_args == config_overrides["ansible_args"]

            # Make sure the playbook exists
            playbook_path = os.path.join(
                config_overrides["ansible_base_path"], "playbook-foo.yaml"
            )
            assert os.path.exists(playbook_path)

            # Read the playbook
            with open(playbook_path, "r") as handle:
                parsed_playbook = yaml.safe_load(handle)
            log.debug3(parsed_playbook)

            # Make sure the playbook params got added
            mod_vars = parsed_playbook[0]["tasks"][0]["vars"]
            for key, val in config_overrides["playbook_parameters"].items():
                assert mod_vars.get(key) == val

            # Make sure the watches file exists
            watches_path = os.path.join(
                config_overrides["ansible_base_path"], "watches.yaml"
            )
            assert os.path.exists(watches_path)

            # Read the watches file
            with open(watches_path, "r") as handle:
                parsed_watches = yaml.safe_load(handle)
            log.debug3(parsed_watches)

            # Make sure the watch entry got the overrides
            assert len(parsed_watches) == 1
            watch = parsed_watches[0]
            assert watch.get("manageStatus") == config_overrides["manage_status"]
            assert (
                watch.get("watchDependentResources")
                == config_overrides["watch_dependent_resources"]
            )
            assert watch.get("reconcilePeriod") == config_overrides["reconcile_period"]


#####################
## Watch/Stop/Wait ##
#####################


@pytest.mark.ansible
def test_watch_start_single_process(workdir):
    """Test that starting multiple managers starts a single process"""

    def test_func():
        while True:
            time.sleep(0.01)

    entrypoint = get_entrypoint(test_func)
    log.debug2(entrypoint)

    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        ansible_entrypoint=entrypoint,
    )

    # Run the watch and make sure the subprocess starts up successfully
    assert wm.watch()
    log.debug("Watch started")
    assert AnsibleWatchManager.ANSIBLE_PROCESS is not None
    assert AnsibleWatchManager.ANSIBLE_PROCESS.poll() is None

    # Stop the process
    wm.stop()
    wm.wait()


@pytest.mark.ansible
def test_watch_environment_passthrough(workdir):
    """Make sure that all external environment variables are passed through to
    the child process when the watch starts
    """

    def test_func():
        assert os.environ.get("_TEST_VAR") == "this is set"

    entrypoint = get_entrypoint(test_func)
    log.debug2(entrypoint)

    os.environ["_TEST_VAR"] = "this is set"
    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        ansible_entrypoint=entrypoint,
    )

    # Run the watch and wait until the process completes
    assert wm.watch()
    wm.wait()

    # Make sure the process terminated cleanly
    assert AnsibleWatchManager.ANSIBLE_PROCESS.returncode == 0


@pytest.mark.ansible
def test_watch_ansible_env(workdir):
    """Make sure that the ansible environment looks right to the spawned process"""

    def test_func():
        log.debug3(os.environ)
        ansible_library = os.environ.get("ANSIBLE_LIBRARY")
        ansible_roles_path = os.environ.get("ANSIBLE_ROLES_PATH")
        assert os.path.exists(ansible_library)
        assert os.path.exists(ansible_roles_path)
        assert "k8s_application.py" in os.listdir(ansible_library)
        assert "oper8_app" in os.listdir(ansible_roles_path)

    entrypoint = get_entrypoint(test_func)
    log.debug2(entrypoint)

    os.environ["_TEST_VAR"] = "this is set"
    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        ansible_entrypoint=entrypoint,
    )

    # Run the watch and wait until the process completes
    assert wm.watch()
    wm.wait()

    # Make sure the process terminated cleanly
    assert AnsibleWatchManager.ANSIBLE_PROCESS.returncode == 0


@pytest.mark.ansible
def test_watch_ansible_args(workdir):
    """Make sure that additional flags passed by `ansible_args` parameter
    are used in the command to spawned process
    """

    def test_func(*args):
        assert args[0] == "--ansible-args"
        assert args[1] == "foo=bar"

    entrypoint = get_entrypoint(test_func)
    log.debug2(entrypoint)

    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        ansible_entrypoint=entrypoint,
        ansible_args="--ansible-args foo=bar ",
    )

    # Run the watch and make sure the subprocess starts up successfully
    assert wm.watch()
    wm.wait()

    # Make sure the assertion in test_func passed
    assert AnsibleWatchManager.ANSIBLE_PROCESS.returncode == 0


@pytest.mark.ansible
def test_watch_subprocess_failure(workdir):
    """Make sure that if the process fails to run, it's handled correctly by the
    watch implementation
    """

    def test_func():
        sys.exit(1)

    entrypoint = get_entrypoint(test_func)
    log.debug2(entrypoint)

    os.environ["_TEST_VAR"] = "this is set"
    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        ansible_entrypoint=entrypoint,
    )

    # Run the watch and wait until the process completes
    # NOTE: The return value of watch is not very useful here because
    #   there's an arbitrary time delta between launching the subprocess and
    #   the process terminating with a non-zero exit code.
    wm.watch()
    wm.wait()

    # Make sure the process terminated cleanly
    assert AnsibleWatchManager.ANSIBLE_PROCESS.returncode == 1


@pytest.mark.ansible
def test_watch_no_new_managers(workdir):
    """Validate that an error is raised if a new manager is created after others
    have been started
    """

    def test_func():
        while True:
            time.sleep(0.01)

    entrypoint = get_entrypoint(test_func)
    log.debug2(entrypoint)

    wm = AnsibleWatchManager(
        DummyController,
        ansible_base_path=workdir,
        ansible_entrypoint=entrypoint,
    )

    # Run the watch and make sure the subprocess starts up successfully
    assert wm.watch()
    log.debug("Watch started")
    assert AnsibleWatchManager.ANSIBLE_PROCESS is not None
    assert AnsibleWatchManager.ANSIBLE_PROCESS.poll() is None

    # Attempt to create a second manager and make sure it errors out
    with pytest.raises(AssertionError):
        AnsibleWatchManager(DummyController)

    # Shut down the running manager
    wm.stop()
    wm.wait()



"""
Base class for all oper8 commands
"""

# Standard
import abc
import argparse


class CmdBase(abc.ABC):
    __doc__ = __doc__

    @abc.abstractmethod
    def add_subparser(
        self,
        subparsers: argparse._SubParsersAction,
    ) -> argparse.ArgumentParser:
        """Add this command's argument parser subcommand

        Args:
            subparsers (argparse._SubParsersAction): The subparser section for
                the central main parser

        Returns:
            subparser (argparse.ArgumentParser): The configured parser for this
                command
        """

    @abc.abstractmethod
    def cmd(self, args: argparse.Namespace):
        """Execute the command with the parsed arguments

        Args:
            args (argparse.Namespace): The parsed command line arguments
        """



"""
The DeployManager is the abstraction in charge of interacting with the
kubernetes cluster to deploy, look up, and delete resources.
"""

# Local
from .base import DeployManagerBase, DeployMethod
from .dry_run_deploy_manager import DryRunDeployManager
from .kube_event import KubeEventType, KubeWatchEvent
from .openshift_deploy_manager import OpenshiftDeployManager



"""Tests for the Reconcile class"""

# Standard
from datetime import datetime, timedelta
from unittest import mock
import copy
import json
import os
import pathlib
import sys
import tempfile

# Third Party
import pytest

# First Party
import aconfig
import alog

# Local
from oper8 import Controller, config, constants, exceptions, status
from oper8.dag.completion_state import CompletionState
from oper8.dag.node import Node
from oper8.deploy_manager import DryRunDeployManager, OpenshiftDeployManager
from oper8.exceptions import ConfigError, RolloutError
from oper8.log_format import Oper8JsonFormatter
from oper8.patch import STRATEGIC_MERGE_PATCH
from oper8.reconcile import ReconcileManager, ReconciliationResult, RequeueParams
from oper8.status import ReadyReason, UpdatingReason
from oper8.test_helpers.helpers import (
    DummyController,
    MockDeployManager,
    library_config,
    make_patch,
    setup_cr,
    setup_session,
    setup_vcs_project,
)
from oper8.vcs import VCSMultiProcessError

log = alog.use_channel("TEST")

################################################################################
## Helpers #####################################################################
################################################################################


class AlogConfigureMock:
    def __init__(self):
        self.kwargs = None

    def __call__(self, **kwargs):
        self.kwargs = kwargs


def check_status(
    deploy_manager, cr, ready_reason=None, updating_reason=None, completion_state=None
):
    """Shared helper for checking status after reconcile"""
    obj = deploy_manager.get_obj(
        kind=cr.kind,
        name=cr.metadata.name,
        namespace=cr.metadata.namespace,
        api_version=cr.apiVersion,
    )
    assert obj is not None
    assert obj.get("status")

    ready_cond = status.get_condition(status.READY_CONDITION, obj["status"])
    if ready_reason:
        assert ready_cond
        assert ready_cond["reason"] == ready_reason.value
    else:
        assert not ready_cond

    update_cond = status.get_condition(status.UPDATING_CONDITION, obj["status"])
    if updating_reason:
        assert update_cond
        assert update_cond["reason"] == updating_reason.value
    else:
        assert not update_cond

    if completion_state:
        expected_component_state = status._make_component_state(completion_state)
        for key, values in expected_component_state.items():
            assert obj.status[status.COMPONENT_STATUS][key] == values


def create_temp_patch_annotation(patch_list=[]):
    combined_patch = {}
    for patch in patch_list:
        name = patch["name"]
        api_version = patch.get("api_version", "org.oper8/v1")
        timestamp = patch.get("timestamp", datetime.now())

        combined_patch[name] = {
            "timestamp": timestamp.isoformat(),
            "api_version": api_version,
        }

    return {constants.TEMPORARY_PATCHES_ANNOTATION_NAME: json.dumps(combined_patch)}


class ReconcileDummyController(DummyController):
    def __init__(self, config_defaults=None):
        super().__init__(
            components=[
                {
                    "name": "foo",
                    "api_objects": [("foo", {"kind": "Foo", "apiVersion": "v1"})],
                },
                {
                    "name": "bar",
                    "api_objects": [("bar", {"kind": "Bar", "apiVersion": "v2"})],
                    "upstreams": ["foo"],
                },
                {
                    "name": "baz",
                    "api_objects": [("baz", {"kind": "Baz", "apiVersion": "v3"})],
                    "disabled": True,
                },
            ],
            config_defaults=config_defaults,
        )

    def finalize_components(self, session):
        session.deploy_manager.deploy(
            [{"apiVersion": "v1", "kind": "Finalized", "metadata": {"name": "test"}}]
        )


# VCS Helpers
@pytest.fixture
def cleanup_vcs():
    dir = os.getcwd()
    path = sys.path

    yield

    # Reset the directory and sys.path
    os.chdir(dir)
    sys.path = path


def create_module_dir(dir):
    if not config.vcs.module_dir:
        return pathlib.Path(dir)

    mod_dir = pathlib.Path(dir) / config.vcs.module_dir
    mod_dir.mkdir(parents=True, exist_ok=True)
    return mod_dir


################################################################################
## Tests #######################################################################
################################################################################

##################
## Construction ##
##################


def test_construct_defaults():
    """Make sure that a reconcile manager can be constructed with its default args"""

    # Basic Construction
    with library_config():
        rm = ReconcileManager()
        assert rm.home_dir == os.getcwd()
        assert rm.vcs == None

    # VCS Enabled
    with library_config(
        vcs={
            "enabled": True,
            "repo": ".",
            "dest": "dest",
            "checkout_method": "worktree",
        }
    ):
        rm = ReconcileManager()
        assert rm.home_dir == "."
        assert rm.vcs is not None

    # Supported Versions Enabled
    with library_config(strict_versioning=True, supported_versions=[]):
        rm = ReconcileManager()


def test_construct_input_args():
    """Make sure that a reconcile manager can be constructed with given args"""
    ReconcileManager(
        home_dir="/tmp/test",
        deploy_manager=MockDeployManager(),
        enable_vcs=False,
    )


def test_construct_exceptions():
    """Make sure that the proper errors are raised by invalid args"""

    # VCS Exceptions
    with library_config(vcs={"enabled": True}):
        with pytest.raises(ConfigError):
            ReconcileManager()

    with library_config(vcs={"enabled": True, "dest": "str"}):
        with pytest.raises(ConfigError):
            ReconcileManager()

    with library_config(
        vcs={
            "enabled": True,
            "dest": "str",
            "repo": "test",
            "checkout_method": "invalid_checkout",
        }
    ):
        with pytest.raises(ConfigError):
            ReconcileManager()

    # Strict Versioning Exceptions
    with library_config(strict_versioning=True, supported_versions=None):
        with pytest.raises(ConfigError):
            ReconcileManager()

    with library_config(strict_versioning=True, supported_versions=["v1"], vcs={}):
        with pytest.raises(ConfigError):
            ReconcileManager()


########################
## parse_manifest ##
########################
@pytest.mark.parametrize(
    ["resource", "result", "raises"],
    [
        [
            {"apiVersion": "v1", "kind": "Pod"},
            aconfig.Config({"apiVersion": "v1", "kind": "Pod"}),
            False,
        ],
        [
            aconfig.Config({"apiVersion": "v1", "kind": "Pod"}),
            aconfig.Config({"apiVersion": "v1", "kind": "Pod"}),
            False,
        ],
        ["BadValue", None, True],
    ],
)
def test_parse_manifest(resource, result, raises):
    """Ensure the ReconcileManager can parse a manifest"""
    if raises:
        with pytest.raises(ValueError):
            ReconcileManager.parse_manifest(resource)
    else:
        manifest = ReconcileManager.parse_manifest(resource)
        assert manifest == result


########################
## configure_logging ##
########################


def test_configure_logging_no_annotations():
    """Make sure that the default logging configuration is applied"""
    rm = ReconcileManager()
    alog_mock = AlogConfigureMock()
    cr = aconfig.Config({})
    with mock.patch("alog.configure", alog_mock):
        rm.configure_logging(cr, "id")

    assert alog_mock.kwargs is not None
    assert alog_mock.kwargs.get("default_level") == "info"
    assert alog_mock.kwargs.get("filters") == ""
    assert alog_mock.kwargs.get("formatter") == "pretty"
    assert alog_mock.kwargs.get("thread_id") is True


def test_configure_logging_custom_json_formatter():
    """Make sure that the if json logging is enabled, the custom formatter is
    used
    """
    rm = ReconcileManager()
    alog_mock = AlogConfigureMock()
    cr = aconfig.Config(
        aconfig.Config(
            {"metadata": {"annotations": {constants.LOG_JSON_NAME: "true"}}}
        ),
        override_env_vars=False,
    )
    with mock.patch("alog.configure", alog_mock):
        rm.configure_logging(cr, "id")
    assert alog_mock.kwargs is not None
    assert alog_mock.kwargs.get("default_level") == "info"
    assert alog_mock.kwargs.get("filters") == ""
    assert isinstance(alog_mock.kwargs.get("formatter"), Oper8JsonFormatter)
    assert alog_mock.kwargs.get("thread_id") is True


def test_configure_logging_with_annotations():
    """Make sure that logging is properly configured if logging annotations given in the
    CR
    """
    rm = ReconcileManager()
    alog_mock = AlogConfigureMock()
    annos = {
        constants.LOG_DEFAULT_LEVEL_NAME: "debug3",
        constants.LOG_FILTERS_NAME: "FOO:debug",
        constants.LOG_JSON_NAME: "false",
        constants.LOG_THREAD_ID_NAME: "true",
    }
    cr = aconfig.Config({"metadata": {"annotations": annos}})
    with mock.patch("alog.configure", alog_mock):
        rm.configure_logging(cr, "id")
    assert alog_mock.kwargs is not None
    assert (
        alog_mock.kwargs.get("default_level") == annos[constants.LOG_DEFAULT_LEVEL_NAME]
    )
    assert alog_mock.kwargs.get("filters") == annos[constants.LOG_FILTERS_NAME]
    assert alog_mock.kwargs.get("formatter") == "pretty"
    assert alog_mock.kwargs.get("thread_id") is True


################################
## generate_id ##
################################


def test_generate_id_uniq():
    """Make sure that two reconciliation IDs don't match"""
    rm = ReconcileManager()
    assert rm.generate_id() != rm.generate_id()


################################
## setup_vcs ##
################################


@pytest.mark.parametrize(
    ["module_dir", "cr"],
    [
        ["src", setup_cr()],
        [None, setup_cr()],
    ],
)
def test_setup_vcs_success_path(cleanup_vcs, module_dir, cr):
    """Test that setup_vcs correctly updates the path/cwd"""

    # Setup required variables
    with tempfile.TemporaryDirectory() as vcs_directory:
        rm = ReconcileManager()

        # Patch setup_directory as that's tested separately and manually create
        # the module directory
        with library_config(vcs={"module_dir": module_dir, "field": "spec.version"}):
            mod_dir = create_module_dir(vcs_directory)
            rm._setup_directory = mock.Mock(return_value=pathlib.Path(vcs_directory))
            rm.setup_vcs(cr)

        # os.getcwd utilizes the full drive path e.g. /private/tmp/... while sys.path is just
        # the human readable /tmp/....
        assert os.getcwd() == str(mod_dir.resolve())
        assert sys.path[0] == str(mod_dir)


def test_setup_vcs_invalid_cr():
    """Ensure setup_vcs raises exceptions on invalid values"""

    # Setup temp directory and the module directory
    rm = ReconcileManager()
    with pytest.raises(ValueError):
        rm.setup_vcs({})


def test_setup_vcs_invalid_module_dir():
    """Ensure setup_vcs raises exception if improper directory setup"""

    # Setup temp directory and cr but not the module directory
    with tempfile.TemporaryDirectory() as vcs_directory:
        cr = setup_cr()

        rm = ReconcileManager()
        rm._setup_directory = mock.Mock(return_value=pathlib.Path(vcs_directory))

        with pytest.raises(ConfigError):
            rm.setup_vcs(cr)


def test_setup_vcs_default_skip_path(cleanup_vcs):
    """Test that setup_vcs correctly skips kind regex matches"""

    # Setup required variables
    with tempfile.TemporaryDirectory() as vcs_directory:
        rm = ReconcileManager()

        # Capture current path/cwd for checks
        current_cwd = os.getcwd()
        current_path = copy.deepcopy(sys.path)

        # Create a TemporaryPatch CR which should be skipped
        cr = setup_cr("TemporaryPatch")

        # Patch setup_directory as that's tested separately and manually create
        # the module directory
        with library_config():
            create_module_dir(vcs_directory)
            rm.setup_vcs(cr)

        # Assert that the cwd is unaffected
        assert os.getcwd() == current_cwd
        assert sys.path == current_path


######################
## setup_controller ##
######################


def test_setup_controller():
    """Test that setup_controller calls its internals correctly"""

    # Setup ReconcileManager with mocks
    rm = ReconcileManager()
    rm._import_controller = mock.Mock(return_value=DummyController)
    rm._configure_controller = mock.Mock(return_value=DummyController())
    rm.setup_controller(DummyController)
    assert rm._import_controller.called
    assert rm._configure_controller.called


##########################
## setup_deploy_manager ##
##########################


@pytest.mark.parametrize(
    ["dry_run", "expected"],
    [
        [True, DryRunDeployManager],
        [False, OpenshiftDeployManager],
    ],
)
def test_setup_deploy_manager(dry_run, expected):
    """Test config settings for deploy manager"""

    rm = ReconcileManager()
    cr = setup_cr()

    with library_config(dry_run=dry_run):
        assert isinstance(rm.setup_deploy_manager(cr), expected)


def test_setup_deploy_manager_override():
    """Ensure the deploy manager passed to the constructor is valid"""

    class TestDM(DryRunDeployManager):
        pass

    dm = TestDM()
    rm = ReconcileManager(deploy_manager=dm)
    cr = setup_cr()

    assert rm.setup_deploy_manager(cr) == dm


################################
## setup_session ##
################################


def test_setup_session():
    """Test setting up a default session"""

    class SessionController(DummyController):
        def get_cr_manifest_defaults(self):
            return aconfig.Config({"spec": {"controllertest": "controllertest"}})

    controller = SessionController()
    cr = setup_cr()
    rm = ReconcileManager()

    rm._get_reconcile_config = mock.Mock(return_value={"test": "test"})
    rm._get_temp_patches = mock.Mock(return_value=[{"test": "test"}])

    session = rm.setup_session(controller, cr, MockDeployManager(), "id")

    assert session.config["test"] == "test"
    assert session.spec["controllertest"] == "controllertest"
    assert session.temporary_patches == [{"test": "test"}]


################################
## run_controller ##
################################


def test_run_controller():
    """Test running a controller"""

    session = setup_session()

    rm = ReconcileManager()
    rm._update_reconcile_start_status = mock.Mock()
    rm._update_reconcile_completion_status = mock.Mock()

    controller = DummyController()
    controller.run_reconcile = mock.Mock(return_value=CompletionState())
    controller.should_requeue = mock.Mock(return_value=(True, None))

    reconcile_result = rm.run_controller(controller, session, is_finalizer=False)

    assert reconcile_result.requeue
    assert reconcile_result.requeue_params.requeue_after == timedelta(minutes=1)
    assert rm._update_reconcile_start_status.called
    assert rm._update_reconcile_completion_status.called


@pytest.mark.parametrize(
    ["should_requeue"],
    [
        [True],
        [False],
    ],
)
def test_run_controller_finalizer(should_requeue):
    """Test running a controller"""

    class FinalizerController(DummyController):
        # Override init to stop overwriting functions
        # with mocks
        def __init__(self):
            Controller.__init__(self)

        def finalize_components(
            self,
            session,
        ):
            pass

        def should_requeue(self, session):
            return should_requeue, None

    session = setup_session()

    rm = ReconcileManager()
    rm._update_reconcile_start_status = mock.Mock()
    rm._update_reconcile_completion_status = mock.Mock()

    controller = FinalizerController()
    controller.run_reconcile = mock.Mock(return_value=CompletionState())
    reconcile_result = rm.run_controller(controller, session, is_finalizer=True)

    assert reconcile_result.requeue == should_requeue
    if should_requeue:
        assert FinalizerController.finalizer in session.finalizers
    else:
        assert FinalizerController.finalizer not in session.finalizers


################################
## _is_paused ##
################################


def test_is_paused():
    """Test checking if a cr is paused"""
    rm = ReconcileManager()
    assert rm._is_paused(
        aconfig.Config(
            {"metadata": {"annotations": {constants.PAUSE_ANNOTATION_NAME: "true"}}}
        )
    )
    assert not rm._is_paused(aconfig.Config({"metadata": {}}))


################################
## _check_strict_versioning ##
################################


def test_check_strict_versioning():
    """Check basic strict versioning functionality"""

    # Create CR, and RM
    cr = setup_cr(version="1.0")
    rm = ReconcileManager()

    # Test valid and invalid paths
    with library_config(supported_versions=["1.0"]):
        rm._check_strict_versioning(cr)

    with library_config(supported_versions=["2.0"]):
        with pytest.raises(ConfigError):
            rm._check_strict_versioning(cr)

    with library_config(supported_versions=["2.0"], vcs={"field": "spec.invalid"}):
        with pytest.raises(ValueError):
            rm._check_strict_versioning(cr)


def test_should_check_strict_versioning():
    """Check basic strict versioning functionality"""

    # Create CR, and RM
    cr = setup_cr(version="1.0")

    # Check global strict_versioning settings
    with library_config(strict_versioning=False):
        rm = ReconcileManager()
        assert not rm._should_check_strict_version(cr)
    with library_config(strict_versioning=True):
        rm = ReconcileManager()
        assert rm._should_check_strict_version(cr)

    # Check various regexs.
    with library_config(
        strict_versioning=True, strict_version_kind_regex=cr.get("kind")
    ):
        rm = ReconcileManager()
        assert rm._should_check_strict_version(cr)
    with library_config(strict_versioning=True, strict_version_kind_regex=".*"):
        rm = ReconcileManager()
        assert rm._should_check_strict_version(cr)
    with library_config(
        strict_versioning=True, strict_version_kind_regex="SomeRandomRegex"
    ):
        rm = ReconcileManager()
        assert not rm._should_check_strict_version(cr)

    # Check default regex for temp patches
    temp_patch_cr = setup_cr("TemporaryPatch")
    with library_config(strict_versioning=True):
        rm = ReconcileManager()
        assert not rm._should_check_strict_version(temp_patch_cr)


def test_check_strict_versioning_vcs():
    """Check strict versioning with VCS"""

    # Create CR, and RM
    cr = setup_cr(version="1.0")
    rm = ReconcileManager()
    rm.vcs = mock.Mock()

    # Test both valid and invalid VCS configurations
    with library_config(supported_versions=["1.0"]):
        rm.vcs.list_refs.return_value = ["1.0"]
        rm._check_strict_versioning(cr)

        rm.vcs.list_refs.return_value = ["2.0"]
        with pytest.raises(ConfigError):
            rm._check_strict_versioning(cr)


################################
## _setup_directory ##
################################


@pytest.mark.parametrize(
    ["cr", "version", "template_dir", "expected_dir", "exception"],
    [
        [setup_cr(version="1.0"), "1.0", "/test/{version}", "/test/1.0", None],
        [setup_cr(version="1.0"), None, "/test/{spec[version]}", "/test/1.0", None],
        [
            setup_cr(),
            "1.0",
            "/test/{version}/{kind}/{name}/",
            "/test/1.0/Widget/test_instance",
            None,
        ],
        [
            setup_cr(deploy_config={"list": [{"test": "expected"}, {"test": "wrong"}]}),
            None,
            "/test/{spec[list][0][test]}",
            "/test/expected",
            None,
        ],
        [setup_cr(version="1.0"), "1.0", "/test/{spec[invalidkey]}", None, ConfigError],
    ],
)
def test_setup_directory(cr, version, template_dir, expected_dir, exception):
    """Check that the setup_directory properly formats the vcs_dir"""
    rm = ReconcileManager()
    rm.vcs = mock.Mock()

    with library_config(vcs={"dest": template_dir, "checkout_method": "worktree"}):
        if exception:
            with pytest.raises(exception):
                rm._setup_directory(cr, version)
        else:
            assert str(rm._setup_directory(cr, version)) == expected_dir


################################
## _import_controller ##
################################


@pytest.mark.parametrize(
    ["controller_info", "exception"],
    [
        [DummyController, None],
        ["oper8.test_helpers.helpers.DummyController", None],
        ["abadcontrollerformat", ConfigError],
        ["oper8.test_helpers.helpers.NotPresentController", ConfigError],
        ["oper8.test_helpers.helpers.FailOnce", ConfigError],
        ["oper8.test_helpers.notpresent.FailOnce", ConfigError],
    ],
)
def test_import_controller(controller_info, exception):
    """Test importing a controller"""

    # Get the module name and expected Controller string from the controller_info
    if isinstance(controller_info, str):
        module_name = controller_info.rsplit(".", 1)[0]
        expected_str = controller_info
    else:
        module_name = controller_info.__module__
        expected_str = f"{controller_info.__module__}.{controller_info.__name__}"

    reimported_module = False
    if module_name in sys.modules:
        sys.modules[module_name].imported = True
        reimported_module = True

    # Import Module or ensure import fails
    rm = ReconcileManager()
    if exception:
        with pytest.raises(exception):
            rm._import_controller(controller_info)
        return

    imported_controller = rm._import_controller(controller_info)

    # Assert the class string is as expected and that the object is actually a controller
    assert issubclass(imported_controller, Controller)
    assert (
        f"{imported_controller.__module__}.{imported_controller.__name__}"
        == expected_str
    )

    if reimported_module:
        assert not hasattr(sys.modules[imported_controller.__module__], "imported")


@pytest.mark.parametrize(
    [
        "reimport_controller",
    ],
    [[True], [False]],
)
def test_reimport_controller(reimport_controller):
    """Test the reimport parameter"""

    rm = ReconcileManager(reimport_controller=reimport_controller)

    # There are some race conditions with all of the reimporting in other tests
    # and fixures, so we explicitly import it here to ensure that it's in
    # sys.modules
    # Local
    from oper8.test_helpers.helpers import DummyController

    # Add attribute to module to test if it was successfully reimported
    sys.modules[DummyController.__module__].imported = True

    imported_controller = rm._import_controller(DummyController)

    # If the module should have been reimported then ensure it doesn't have the "imported"
    # attribute
    if reimport_controller:
        assert not hasattr(sys.modules[imported_controller.__module__], "imported")
    else:
        assert sys.modules[imported_controller.__module__].imported


def test_unimport_controller_class():
    """Test that _unimport_controller_class correctly removes all parents,
    siblings, children, and cousins from sys.modules
    """
    mod_name = "foo.bar"
    mods = {
        # Self
        mod_name: 1,
        # Parent
        "foo": 2,
        # Child
        "foo.bar.bif": 3,
        # Sibling
        "foo.baz": 4,
        # Cousin
        "foo.baz.bonk": 5,
        # No relation!
        "asdf": 6,
    }
    with mock.patch("sys.modules", new=copy.copy(mods)):
        unimported_modules = ReconcileManager._unimport_controller_module(mod_name)
        assert unimported_modules == {mod for mod in mods if mod.startswith("foo")}
        assert "asdf" in sys.modules
        assert not any(mod in sys.modules for mod in unimported_modules)


################################
## _configure_controller ##
################################


def test_configure_controller():
    """Test configuring a normal controller"""
    rm = ReconcileManager()
    controller = rm._configure_controller(DummyController)
    assert isinstance(controller, DummyController)


###########################
## _get_reconcile_config ##
###########################


example_reconcile_config = {"foo": 1, "bar": {"baz": 2}}
example_reconcile_config_map = {
    "kind": "ConfigMap",
    "apiVersion": "v1",
    "metadata": {"name": "config-cm"},
    "data": example_reconcile_config,
}


@pytest.mark.parametrize(
    ["cr_manifest", "controller_defaults", "resources", "expected_config", "exception"],
    [
        # Test config with defaults, overrides, or both
        [aconfig.Config({}), aconfig.Config({}), [], aconfig.Config({}), None],
        [
            aconfig.Config({constants.CONFIG_OVERRIDES: example_reconcile_config}),
            aconfig.Config({}),
            [],
            aconfig.Config(example_reconcile_config),
            None,
        ],
        [
            aconfig.Config({}),
            aconfig.Config(example_reconcile_config),
            [],
            aconfig.Config(example_reconcile_config),
            None,
        ],
        [
            aconfig.Config({constants.CONFIG_OVERRIDES: {"foo": 3}}),
            aconfig.Config(example_reconcile_config),
            [],
            aconfig.Config({"foo": 3, "bar": {"baz": 2}}),
            None,
        ],
        # Test get config from configmaps
        [
            aconfig.Config(
                {
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    }
                }
            ),
            aconfig.Config({}),
            [example_reconcile_config_map],
            aconfig.Config(example_reconcile_config),
            None,
        ],
        [
            aconfig.Config(
                {
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    }
                }
            ),
            aconfig.Config({"foo": 3}),
            [example_reconcile_config_map],
            aconfig.Config(example_reconcile_config),
            None,
        ],
        [
            aconfig.Config(
                {
                    constants.CONFIG_OVERRIDES: {"foo": 3},
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    },
                }
            ),
            aconfig.Config({"foo": 0}),
            [example_reconcile_config_map],
            aconfig.Config({"foo": 3, "bar": {"baz": 2}}),
            None,
        ],
        [
            aconfig.Config(
                {
                    constants.CONFIG_OVERRIDES: {"foo": 3},
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    },
                }
            ),
            aconfig.Config({"baz": 0, "foo": 0}),
            [example_reconcile_config_map],
            aconfig.Config({"foo": 3, "baz": 0, "bar": {"baz": 2}}),
            None,
        ],
        # Test _get_reconcile config exceptions
        [
            aconfig.Config(
                {
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    }
                }
            ),
            aconfig.Config({}),
            [],
            aconfig.Config({}),
            ConfigError,
        ],
        [
            aconfig.Config(
                {
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    }
                }
            ),
            aconfig.Config({}),
            [
                {
                    "kind": "ConfigMap",
                    "apiVersion": "v1",
                    "metadata": {"name": "config-cm"},
                }
            ],
            aconfig.Config({}),
            ConfigError,
        ],
        [
            aconfig.Config(
                {
                    "metadata": {
                        "annotations": {
                            constants.CONFIG_DEFAULTS_ANNOTATION_NAME: "config-cm"
                        }
                    }
                }
            ),
            aconfig.Config({}),
            [
                {
                    "kind": "ConfigMap",
                    "apiVersion": "v1",
                    "metadata": {"name": "config-cm"},
                    "data": 123,
                }
            ],
            aconfig.Config({}),
            ConfigError,
        ],
    ],
)
def test_get_reconcile_config(
    cr_manifest, controller_defaults, resources, expected_config, exception
):
    """Test all permutations of generating the reconcile config"""
    dm = MockDeployManager(resources=resources)
    rm = ReconcileManager(deploy_manager=dm)

    if exception:
        with pytest.raises(exception):
            rm._get_reconcile_config(cr_manifest, dm, controller_defaults)
        return

    assert (
        rm._get_reconcile_config(cr_manifest, dm, controller_defaults)
        == expected_config
    )


#######################
## _get_temp_patches ##
#######################


@pytest.mark.parametrize(
    ["cr_manifest", "patches", "exception"],
    [
        # Test CR without patches
        [setup_cr(), [], None],
        # Test generic patch
        [
            setup_cr(
                metadata={
                    "annotations": create_temp_patch_annotation([{"name": "test"}])
                }
            ),
            [
                make_patch(
                    patch_type=STRATEGIC_MERGE_PATCH,
                    body={"foo": {"new": "value"}},
                    name="test",
                    target=setup_cr(),
                )
            ],
            None,
        ],
        # Test patches are parsed in the right order
        [
            setup_cr(
                metadata={
                    "annotations": create_temp_patch_annotation(
                        [
                            {
                                "name": "test-newer",
                                "timestamp": datetime.now() + timedelta(minutes=1),
                            },
                            {"name": "test"},
                        ]
                    )
                }
            ),
            [
                make_patch(
                    patch_type=STRATEGIC_MERGE_PATCH,
                    body={"foo": {"new": "value"}},
                    name="test",
                    target=setup_cr(),
                ),
                make_patch(
                    patch_type=STRATEGIC_MERGE_PATCH,
                    body={"foo": {"new": "value"}},
                    name="test-newer",
                    target=setup_cr(),
                ),
            ],
            None,
        ],
        # Test non json patch
        [
            setup_cr(
                metadata={
                    "annotations": {
                        constants.TEMPORARY_PATCHES_ANNOTATION_NAME: "notjson"
                    }
                }
            ),
            [],
            RolloutError,
        ],
        # Test non dict patch
        [
            setup_cr(
                metadata={
                    "annotations": {
                        constants.TEMPORARY_PATCHES_ANNOTATION_NAME: '[{"not": "a dict"}]'
                    }
                }
            ),
            [],
            RolloutError,
        ],
        # Test invalid timestamp
        [
            setup_cr(
                metadata={
                    "annotations": {
                        constants.TEMPORARY_PATCHES_ANNOTATION_NAME: '{"patch":{"api_version":"org.oper8/v1","timestamp":"invalid"}}'
                    }
                }
            ),
            [],
            RolloutError,
        ],
        # Test invalid api_version
        [
            setup_cr(
                metadata={
                    "annotations": {
                        constants.TEMPORARY_PATCHES_ANNOTATION_NAME: '{"patch":{"timestamp":"2000-01-01T00:00:00"}}'
                    }
                }
            ),
            [],
            RolloutError,
        ],
        # Test temporary patch not found
        [
            setup_cr(
                metadata={
                    "annotations": create_temp_patch_annotation([{"name": "test"}])
                }
            ),
            [],
            ConfigError,
        ],
        [
            setup_cr(
                metadata={
                    "annotations": create_temp_patch_annotation([{"name": "test"}])
                }
            ),
            [
                {
                    **make_patch(
                        patch_type=STRATEGIC_MERGE_PATCH,
                        body={"foo": {"new": "value"}},
                        name="test",
                        target=setup_cr(),
                    ),
                    # Forcibly remove spec from parameter
                    "spec": None,
                }
            ],
            ConfigError,
        ],
    ],
)
def test_get_temp_patches(cr_manifest, patches, exception):
    """Test getting temporary patches"""
    dm = MockDeployManager(resources=patches)
    rm = ReconcileManager(deploy_manager=dm)

    if exception:
        with pytest.raises(exception):
            rm._get_temp_patches(dm, cr_manifest)
        return

    gathered_patches = rm._get_temp_patches(dm, cr_manifest)
    assert len(gathered_patches) == len(patches)
    for index in range(len(patches)):
        assert gathered_patches[index] == patches[index]


#############################
## _update_resource_status ##
#############################


def test_update_resource_status():
    """Test updating resource status"""
    dm = MockDeployManager()
    the_api_version = "my.api/v123alpha0"
    cr = setup_cr(api_version=the_api_version)
    rm = ReconcileManager(deploy_manager=dm)
    with mock.patch("oper8.status.update_resource_status") as Mock:
        rm._update_resource_status(dm, cr, current_status={"test": "status"})
        Mock.assert_called_with(
            dm,
            cr.kind,
            the_api_version,
            cr.metadata.name,
            cr.metadata.namespace,
            current_status={"test": "status"},
        )


####################################
## _update_reconcile_start_status ##
####################################


@pytest.mark.parametrize(
    ["cr", "ready_reason", "updating_reason"],
    [
        # Test initializing condition
        [setup_cr(), ReadyReason.INITIALIZING, None],
        # Test version update conditions
        [
            setup_cr(
                version="2.0",
                status={
                    "conditions": [
                        status._make_ready_condition(
                            ReadyReason.STABLE, "", datetime.now()
                        )
                    ],
                    "versions": {"reconciled": "1.0"},
                },
            ),
            ReadyReason.IN_PROGRESS,
            UpdatingReason.VERSION_CHANGE,
        ],
        # Test that the RM passes through conditions if not version_change/initializing
        [
            setup_cr(
                version="1.0",
                status={
                    "conditions": [
                        status._make_ready_condition(
                            ReadyReason.STABLE, "", datetime.now()
                        ),
                        status._make_updating_condition(
                            UpdatingReason.STABLE, "", datetime.now()
                        ),
                    ],
                    "versions": {"reconciled": "1.0"},
                },
            ),
            ReadyReason.STABLE,
            UpdatingReason.STABLE,
        ],
        [
            setup_cr(
                version="1.0",
                status={
                    "conditions": [
                        status._make_ready_condition(
                            ReadyReason.CONFIG_ERROR, "", datetime.now()
                        ),
                        status._make_updating_condition(
                            UpdatingReason.ERRORED, "", datetime.now()
                        ),
                    ],
                    "versions": {"reconciled": "1.0"},
                },
            ),
            ReadyReason.CONFIG_ERROR,
            UpdatingReason.ERRORED,
        ],
    ],
)
def test_update_reconcile_start_status(cr, ready_reason, updating_reason):
    """Test correctly updating reconcile start status"""
    dm = MockDeployManager(resources=[cr])
    rm = ReconcileManager(deploy_manager=dm)
    session = setup_session(full_cr=cr, deploy_manager=dm)

    rm._update_reconcile_start_status(session)

    check_status(dm, cr, ready_reason, updating_reason)


###############
## _update_reconcile_completion_status ##
###############


@pytest.mark.parametrize(
    ["cr", "completion_state", "ready_reason", "updating_reason"],
    [
        # Test stable on successful reconcile
        [
            setup_cr(),
            CompletionState(verified_nodes=[Node("test")]),
            ReadyReason.STABLE,
            UpdatingReason.STABLE,
        ],
        # Test conditions on incomplete verification
        [
            setup_cr(),
            CompletionState(unverified_nodes=[Node("test")]),
            ReadyReason.IN_PROGRESS,
            UpdatingReason.VERIFY_WAIT,
        ],
        [
            setup_cr(
                status={
                    "conditions": [
                        status._make_ready_condition(
                            ReadyReason.INITIALIZING, "", datetime.now()
                        )
                    ]
                }
            ),
            CompletionState(unverified_nodes=[Node("test")]),
            ReadyReason.INITIALIZING,
            UpdatingReason.VERIFY_WAIT,
        ],
    ],
)
def test_update_reconcile_completion_status(
    cr, completion_state, ready_reason, updating_reason
):
    """Test updating a resource status with a completion state"""
    dm = MockDeployManager(resources=[cr])
    rm = ReconcileManager(deploy_manager=dm)
    session = setup_session(full_cr=cr, deploy_manager=dm)

    rm._update_reconcile_completion_status(session, completion_state)

    check_status(dm, cr, ready_reason, updating_reason, completion_state)


###############
## _update_error_status ##
###############


@pytest.mark.parametrize(
    ["cr", "exception", "ready_reason", "updating_reason"],
    [
        # Test stable on successful reconcile
        [
            setup_cr(),
            exceptions.PreconditionError(),
            None,
            UpdatingReason.PRECONDITION_WAIT,
        ],
        [setup_cr(), exceptions.VerificationError(), None, UpdatingReason.VERIFY_WAIT],
        [setup_cr(), exceptions.Oper8ExpectedError(), None, UpdatingReason.VERIFY_WAIT],
        [
            setup_cr(),
            exceptions.ConfigError(),
            ReadyReason.CONFIG_ERROR,
            UpdatingReason.ERRORED,
        ],
        [setup_cr(), exceptions.ClusterError(), None, UpdatingReason.CLUSTER_ERROR],
        [
            setup_cr(),
            exceptions.RolloutError(),
            ReadyReason.ERRORED,
            UpdatingReason.ERRORED,
        ],
        [
            setup_cr(),
            exceptions.Oper8FatalError(),
            ReadyReason.ERRORED,
            UpdatingReason.ERRORED,
        ],
        [setup_cr(), ValueError(), ReadyReason.ERRORED, UpdatingReason.ERRORED],
    ],
)
def test_update_error_status(cr, exception, ready_reason, updating_reason):
    """Test updating the CR status on error"""
    dm = MockDeployManager(resources=[cr])
    rm = ReconcileManager(deploy_manager=dm)
    rm._update_error_status(cr, exception)
    check_status(dm, cr, ready_reason, updating_reason)


###############
## reconcile ##
###############


@pytest.mark.parametrize(
    ["controller_info", "cr", "is_finalizer"],
    [
        # Test controller_info as both strings and type
        [ReconcileDummyController, setup_cr(), False],
        [
            f"{ReconcileDummyController.__module__}.{ReconcileDummyController.__name__}",
            setup_cr(),
            False,
        ],
        # Test finalizer
        [
            ReconcileDummyController,
            setup_cr(
                metadata={
                    "finalizers": [ReconcileDummyController.finalizer],
                    "deletionTimestamp": datetime.now(),
                }
            ),
            True,
        ],
    ],
)
def test_reconcile(controller_info, cr, is_finalizer):
    """Test that a reconciliation with several components that complete cleanly
    exists correctly and sets status conditions to STABLE
    """
    dm = MockDeployManager(resources=[cr])
    rm = ReconcileManager(deploy_manager=dm)

    result = rm.reconcile(controller_info, cr, is_finalizer)
    assert not result.requeue

    if is_finalizer:
        assert dm.has_obj(kind="Finalized", name="test", api_version="v1")
        assert not dm.has_obj(
            kind=cr.kind,
            api_version=cr.apiVersion,
            name=cr.metadata.name,
            namespace=cr.metadata.namespace,
        )
    else:
        assert dm.has_obj(
            kind="Foo", name="foo", api_version="v1", namespace=cr.metadata.namespace
        )
        assert dm.has_obj(
            kind="Bar", name="bar", api_version="v2", namespace=cr.metadata.namespace
        )

        # make sure status gets STABLE
        #   and requeue won't be required if status reaches STABLE
        check_status(dm, cr, ReadyReason.STABLE, UpdatingReason.STABLE)


def test_reconcile_paused():
    """Reconcile is not ran for a paused CR"""
    cr = setup_cr(metadata={"annotations": {constants.PAUSE_ANNOTATION_NAME: "true"}})
    dm = MockDeployManager(resources=[cr])
    rm = ReconcileManager(deploy_manager=dm)
    rm.run_controller = mock.Mock()

    result = rm.reconcile(ReconcileDummyController, cr, False)

    assert result == ReconciliationResult(False, RequeueParams())
    assert not rm.run_controller.called


@pytest.mark.parametrize(
    ["cr", "supported_versions", "exception"],
    [
        [setup_cr(version="1.0"), ["1.0"], None],
        [setup_cr(version="2.0"), ["1.0"], ConfigError],
    ],
)
def test_reconcile_strict_versioning(cr, supported_versions, exception):
    """Test reconcile is stopped by strict versioning"""

    with library_config(strict_versioning=True, supported_versions=supported_versions):
        dm = MockDeployManager(resources=[cr])
        rm = ReconcileManager(deploy_manager=dm)

        if exception:
            with pytest.raises(exception):
                rm.reconcile(ReconcileDummyController, cr)
        else:
            rm.reconcile(ReconcileDummyController, cr)


@pytest.mark.parametrize(
    ["module_dir"],
    [["spec"], ["adifferentdir"]],
)
def test_reconcile_vcs(cleanup_vcs, module_dir):
    """Full VCS Integration Test utilizing a temporary git directory"""

    with tempfile.TemporaryDirectory() as vcs_directory:
        checkout_path = f"{vcs_directory}/versions/{{version}}"
        setup_vcs_project(
            components=ReconcileDummyController().component_specs,
            directory=vcs_directory,
            module_dir=module_dir,
            module_name="test_module",
        )

        cr = setup_cr()

        with library_config(
            vcs={
                "enabled": True,
                "field": "spec.version",
                "repo": vcs_directory,
                "dest": checkout_path,
                "module_dir": module_dir,
                "checkout_method": "worktree",
            }
        ):
            dm = MockDeployManager(resources=[cr])
            rm = ReconcileManager(deploy_manager=dm)

            result = rm.reconcile("test_module.Controller", cr)

        assert dm.has_obj(
            kind="Foo", name="foo", api_version="v1", namespace=cr.metadata.namespace
        )
        assert dm.has_obj(
            kind="Bar", name="bar", api_version="v2", namespace=cr.metadata.namespace
        )

        # make sure status gets STABLE
        #   and requeue won't be required if status reaches STABLE
        check_status(dm, cr, ReadyReason.STABLE, UpdatingReason.STABLE)
        assert not result.requeue


####################
## Safe Reconcile ##
####################


def test_safe_reconcile():
    """Test that a safe_reconcile handles all exception types"""
    cr = setup_cr()
    rm = ReconcileManager()
    rm.reconcile = mock.Mock()

    with library_config(manage_status=False):
        # Test a successful reconcile
        good_result = ReconciliationResult(False, RequeueParams())
        rm.reconcile.return_value = good_result
        assert good_result == rm.safe_reconcile(DummyController, cr)

        # Test a VCS error
        rm.reconcile.side_effect = VCSMultiProcessError()
        result = rm.safe_reconcile(DummyController, cr)
        assert result.requeue
        assert result.requeue_params.requeue_after > timedelta(
            seconds=5
        ) and result.requeue_params.requeue_after < timedelta(seconds=10)

        # Test a generic exception
        rm.reconcile.side_effect = RolloutError()
        result = rm.safe_reconcile(DummyController, cr)
        assert result.requeue
        assert result.requeue_params.requeue_after == timedelta(seconds=60)


def test_safe_reconcile_status():
    """Test a safe reconcile correctly updates the error status"""
    cr = setup_cr()
    dm = MockDeployManager(resources=[cr])
    rm = ReconcileManager(deploy_manager=dm)
    rm.reconcile = mock.Mock(side_effect=RolloutError())
    rm._update_error_status = mock.Mock()

    # Test successful status update
    assert rm.safe_reconcile(DummyController, cr)

    # Test safe reconcile when error in updating status
    rm._update_error_status.side_effect = RolloutError()
    assert rm.safe_reconcile(DummyController, cr)


#################
## Data models ##
#################


def test_create_requeue_params_default():
    """Make sure RequeueParams can be constructed by empty args"""
    p = RequeueParams()
    assert p.requeue_after == timedelta(seconds=60)

    with library_config(requeue_after_seconds=1):
        p = RequeueParams()
        assert p.requeue_after == timedelta(seconds=1)


def test_create_requeue_params_args():
    """Make sure requeue parameters can be constructed with args"""
    p = RequeueParams(requeue_after=timedelta(seconds=100))
    assert p.requeue_after == timedelta(seconds=100)


def test_create_reconcile_request_defaults():
    """Make sure reconcile request can be constructed by empty args"""
    rr = ReconciliationResult(True)
    assert rr.requeue
    assert rr.requeue_params.requeue_after == timedelta(seconds=60)
    assert rr.exception is None

    with library_config(requeue_after_seconds=1):
        rr = ReconciliationResult(False)

    assert not rr.requeue
    assert rr.requeue_params.requeue_after == timedelta(seconds=1)
    assert rr.exception is None


def test_create_reconcile_request_args():
    """Make sure reconcile request can be constructed with args"""
    p = RequeueParams(requeue_after=timedelta(seconds=100))
    rr = ReconciliationResult(True, p, ValueError())

    assert rr.requeue
    assert rr.requeue_params.requeue_after == timedelta(seconds=100)
    assert isinstance(rr.exception, ValueError)



"""
This module just loads config at import time and does the initial log config
"""

# Standard
import os

# First Party
import aconfig
import alog

# Local
from .validation import get_invalid_params

# Read the library config, allowing env overrides
library_config = aconfig.Config.from_yaml(
    os.path.join(os.path.dirname(__file__), "config.yaml"),
    override_env_vars=True,
)

# Parse the validation file, not allowing env overrides
validation_config = aconfig.Config.from_yaml(
    os.path.join(os.path.dirname(__file__), "config_validation.yaml"),
    override_env_vars=False,
)

# Validate the loaded config values
invalid_params = get_invalid_params(library_config, validation_config)
assert (
    not invalid_params
), f"Library configuration found invalid values: {invalid_params}"

# Do initial alog configuration
alog.configure(
    default_level=library_config.log_level,
    filters=library_config.log_filters,
    formatter="json" if library_config.log_json else "pretty",
    thread_id=library_config.log_thread_id,
)



"""
Base class definition for all datastore connections
"""

# Standard
from typing import Optional
import abc

# First Party
import alog

# Local
from ... import Session, assert_cluster
from ..utils import common
from ..utils.abc_static import ABCStatic

log = alog.use_channel("DCONN")


class DatastoreConnectionBase(ABCStatic):
    """
    A DatastoreConnection is an object that holds all of the critical data to
    connect to a specific datastore type. A DatastoreConnection for a given
    datastore type MUST not care what implementation backs the connection.
    """

    ## Construction ############################################################

    def __init__(self, session: Session):
        """Construct with the session so that it can be saved as a member"""
        self._session = session

    @property
    def session(self) -> Session:
        return self._session

    ## Abstract Interface ######################################################

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Serialize the internal connection details to a dict object which can
        be added directly to a subsystem's CR.

        Returns:
            config_dict:  dict
                This dict will hold the keys and values that can be used to add
                to a subsystem's datastores.connections section.
        """

    @classmethod
    @abc.abstractmethod
    def from_dict(
        cls, session: Session, config_dict: dict
    ) -> "DatastoreConnectionBase":
        """Parse a config_dict from a subsystem CR to create an instance of the
        DatastoreConnection class.

        Args:
            session:  Session
                The current deploy session
            config_dict:  dict
                This dict will hold the keys and values created by to_dict and
                pulled from the subsystem CR.

        Returns:
            datastore_connection:  DatastoreConnectionBase
                The constructed instance of the connection
        """

    ## Shared Utilities ########################################################

    def _fetch_secret_data(self, secret_name: str) -> Optional[dict]:
        """Most connection implementations will need the ability to fetch secret
        data from the cluster when loading from the CR dict, so this provides a
        common implementation.

        Args:
            secret_name:  str
                The name of the secret to fetch

        Returns:
            secret_data:  Optional[dict]
                The content of the 'data' field in the secret with values base64
                decoded if the secret is found, otherwise None
        """
        success, content = self.session.get_object_current_state("Secret", secret_name)
        assert_cluster(success, f"Fetching connection secret [{secret_name}] failed")
        if content is None:
            return None
        assert "data" in content, "Got a secret without 'data'?"
        return {
            key: common.b64_secret_decode(val) for key, val in content["data"].items()
        }



"""
This module adds metaclass support for declaring an interface with
@abstractmethod methods that MUST be implemented as @classmethod or
@staticmethod
"""

# Standard
import abc
import inspect


class ABCStaticMeta(abc.ABCMeta):
    """The StaticABCMeta class is a metaclass that enforces implementations of
    base class functions marked as both @abstractmethod and @classmethod.
    Methods with this signature MUST be implemented with the @classmethod or
    @staticmethod decorator in derived classes.
    """

    def __init__(cls, name, bases, dct):
        # Find abstract class methods that have not been implemented at all
        attrs = {name: getattr(cls, name) for name in dir(cls)}
        cls.__abstract_class_methods__ = [
            name
            for name, attr in attrs.items()
            if inspect.ismethod(attr) and getattr(attr, "__isabstractmethod__", False)
        ]

        # For any abstract class methods that have not been implemented,
        # overwrite them to raise NotImplementedError if called
        for method_name in cls.__abstract_class_methods__:

            def not_implemented(*_, x=method_name, **__):
                raise NotImplementedError(f"Cannot invoke abstract class method {x}")

            not_implemented.__original_signature__ = inspect.signature(
                getattr(cls, method_name)
            )
            setattr(cls, method_name, not_implemented)

        # Look for abstract class methods of parents
        base_abstract_class_methods = {
            method_name: getattr(base, method_name)
            for base in bases
            for method_name in getattr(base, "__abstract_class_methods__", [])
            if method_name not in cls.__abstract_class_methods__
        }

        # If any parent abstract class methods have been implemented as instance
        # methods, raise an import-time exception
        for method_name, base_method in base_abstract_class_methods.items():
            # A local implementation is valid if it is a bound method (
            # implemented as a @classmethod) or it is a function with a
            # signature that exactly matches the signature of the base class
            # (implemented as @staticmethod).
            this_method = getattr(cls, method_name)
            is_classmethod = inspect.ismethod(this_method)
            original_signature = getattr(base_method, "__original_signature__", None)
            is_staticmethod = inspect.isfunction(this_method) and inspect.signature(
                this_method
            ) in [original_signature, inspect.signature(base_method)]
            if not (is_classmethod or is_staticmethod):
                raise NotImplementedError(
                    f"The method [{method_name}] is an @classmethod @abstractmethod. "
                    f"{cls} implements it as an instance method"
                )


class ABCStatic(metaclass=ABCStaticMeta):
    """An ABCStatic class is a child of abc.ABC which has support for enforcing
    methods which combine @classmethod and @abstractmethod
    """



"""
Redis instance factory
"""

# Local
from ..factory_base import DatastoreSingletonFactoryBase
from .connection import RedisConnection


class RedisFactory(DatastoreSingletonFactoryBase):
    """The common factory that will manage instances of Redis"""

    DATASTORE_TYPE = "redis"
    CONNECTION_TYPE = RedisConnection



"""
This module holds the base class interface for the various implementations of
WatchManager
"""

# Standard
from typing import Type
import abc

# First Party
import alog

# Local
from ..controller import Controller

log = alog.use_channel("WATCH")


class WatchManagerBase(abc.ABC):
    """A WatchManager is responsible for linking a kubernetes custom resource
    type with a Controller that will execute the reconciliation loop
    """

    # Class-global mapping of all watches managed by this operator
    _ALL_WATCHES = {}

    ## Interface ###############################################################

    def __init__(
        self,
        controller_type: Type[Controller],
    ):
        """Construct with the controller type that will be watched

        Args:
            controller_type:  Type[Controller],
                The Controller instance that will manage this group/version/kind
        """
        self.controller_type = controller_type
        self.group = controller_type.group
        self.version = controller_type.version
        self.kind = controller_type.kind

        # Register this watch instance
        watch_key = str(self)
        assert (
            watch_key not in self._ALL_WATCHES
        ), "Only a single controller may watch a given group/version/kind"
        self._ALL_WATCHES[watch_key] = self

    @abc.abstractmethod
    def watch(self) -> bool:
        """The watch function is responsible for initializing the persistent
        watch and returning whether or not the watch was started successfully.

        Returns:
            success:  bool
                True if the watch was spawned correctly, False otherwise.
        """

    @abc.abstractmethod
    def wait(self):
        """The wait function is responsible for blocking until the managed watch
        has been terminated.
        """

    @abc.abstractmethod
    def stop(self):
        """Terminate this watch if it is currently running"""

    ## Utilities ###############################################################

    @classmethod
    def start_all(cls) -> bool:
        """This utility starts all registered watches

        Returns:
            success:  bool
                True if all watches started succssfully, False otherwise
        """
        started_watches = []
        success = True
        # NOTE: sorting gives deterministic order so that launch failures can be
        #   diagnosed (and tested) more easily. This is not strictly necessary,
        #   but it also doesn't hurt and it is nice to have.
        for _, watch in sorted(cls._ALL_WATCHES.items()):
            if watch.watch():
                log.debug("Successfully started %s", watch)
                started_watches.append(watch)
            else:
                log.warning("Failed to start %s", watch)
                success = False

                # Shut down all successfully started watches
                for started_watch in started_watches:
                    started_watch.stop()

                # Don't start any of the others
                break

        # Wait on all of them to terminate
        for watch in cls._ALL_WATCHES.values():
            watch.wait()

        return success

    @classmethod
    def stop_all(cls):
        """This utility stops all watches"""
        for watch in cls._ALL_WATCHES.values():
            try:
                watch.stop()
                log.debug2("Waiting for %s to terminate", watch)
                watch.wait()
            except Exception as exc:  # pylint: disable=broad-exception-caught
                log.error("Failed to stop watch manager %s", exc, exc_info=True)

    ## Implementation Details ##################################################

    def __str__(self):
        """String representation of this watch"""
        return f"Watch[{self.controller_type}]"



"""
Thread class that will dump a heartbeat to a file periodically
"""

# Standard
from datetime import datetime
import threading

# First Party
import alog

# Local
from ..utils import parse_time_delta
from .timer import TimerThread

log = alog.use_channel("HBEAT")


class HeartbeatThread(TimerThread):
    """The HeartbeatThread acts as a pulse for the PythonWatchManager.

    This thread will periodically dump the value of "now" to a file which can be
    read by an observer such as a liveness/readiness probe to ensure that the
    manager is functioning well.
    """

    # This format is designed to be read using `date -d $(cat heartbeat.txt)`
    # using the GNU date utility
    # CITE: https://www.gnu.org/software/coreutils/manual/html_node/Examples-of-date.html
    _DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    def __init__(self, heartbeat_file: str, heartbeat_period: str):
        """Initialize with the file location for the heartbeat output

        Args:
            heartbeat_file: str
                The fully-qualified path to the heartbeat file
            heartbeat_period: str
                Time delta string representing period delay between beats.
                NOTE: The GNU `date` utility cannot parse sub-seconds easily, so
                    the expected configuration for this is to be >= 1s
        """
        self._heartbeat_file = heartbeat_file
        self._offset = parse_time_delta(heartbeat_period)
        self._beat_lock = threading.Lock()
        self._beat_event = threading.Event()
        super().__init__(name="heartbeat_thread")

    def run(self):
        self._run_heartbeat()
        return super().run()

    def wait_for_beat(self):
        """Wait for the next beat"""
        # Make sure the beat lock is not held before starting wait. This
        # prevents beats that are immediately ready
        with self._beat_lock:
            pass

        # Wait for the next beat
        self._beat_event.wait()

    def _run_heartbeat(self):
        """Run the heartbeat dump to the heartbeat file and put the next beat"""
        now = datetime.now()
        log.debug3("Heartbeat %s", now)

        # Save the beat to disk
        try:
            with open(self._heartbeat_file, "w", encoding="utf-8") as handle:
                handle.write(now.strftime(self._DATE_FORMAT))
                handle.flush()
        except Exception as err:
            log.warning("Failed to write heartbeat file: %s", err, exc_info=True)

        # Unblock and reset the wait condition
        with self._beat_lock:
            self._beat_event.set()
            self._beat_event.clear()

        # Put the next beat if not stopped
        if not self.should_stop():
            self.put_event(now + self._offset, self._run_heartbeat)



"""
The common Connection type for a postgres instance
"""

# Standard
from typing import List, Optional, Tuple

# First Party
import alog

# Local
from .... import Session, assert_cluster, assert_config, assert_precondition
from ...utils import common
from .. import DatastoreConnectionBase

DEFAULT_TLS_VOLUME_NAME = "postgres-tls"
DEFAULT_TLS_VOLUME_MOUNT_PATH = "/tls"


log = alog.use_channel("PGCON")


class PostgresConnection(DatastoreConnectionBase):
    """A connection for postgres defines the client operations and utilities
    needed to configure a microservice to interact with a single postgres
    instance. The key pieces of information are:

    * General config:
        * hostname: The hostname where the database can be reached
        * port: The port the database service is listening on

    * Auth:
        * auth_secret_name: The in-cluster name for the secret holding the
            username and password
        * auth_secret_username_field: The field within the auth secret that
            holds the username
        * auth_secret_password_field: The field within the auth secret that
            holds the password

    * TLS:
        * tls_cert: The content of the TLS cert if tls is enabled
        * tls_secret_name: The in-cluster name for the secret holding the TLS
            creds if tls is enabled
        * tls_secret_cert_field: The field within the tls secret that holds the
            cert
    """

    def __init__(
        self,
        session: Session,
        hostname: str,
        port: int,
        auth_secret_name: str,
        auth_secret_username_field: str,
        auth_secret_password_field: str,
        tls_secret_name: Optional[str] = None,
        tls_secret_cert_field: Optional[str] = None,
        auth_username: Optional[str] = None,
        auth_password: Optional[str] = None,
        tls_cert: Optional[str] = None,
    ):
        """Construct with all of the crucial information pieces"""
        super().__init__(session)

        # Save internal values
        self._hostname = hostname
        self._port = port
        self._auth_secret_name = auth_secret_name
        self._auth_secret_username_field = auth_secret_username_field
        self._auth_secret_password_field = auth_secret_password_field
        self._tls_secret_name = tls_secret_name
        self._tls_secret_cert_field = tls_secret_cert_field

        # The secret content may be populated or not, depending on whether this
        # Connection is being created by the component or a CR config. If not
        # populated now, they will be lazily populated on client request.
        self._auth_username = auth_username
        self._auth_password = auth_password
        self._tls_cert = tls_cert

        # Ensure that the TLS arguments are provided in a reasonable way. The
        # cert may be omitted
        tls_args = {tls_secret_name, tls_secret_cert_field}
        assert (
            tls_args == {None} or None not in tls_args
        ), "All TLS arguments must be provided if tls is enabled"
        self._tls_enabled = tls_args != {None}
        assert (
            self._tls_enabled or self._tls_cert is None
        ), "Cannot give a tls cert value when tls is disabled"

    ## Properties ##############################################################

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def port(self) -> int:
        return self._port

    @property
    def auth_secret_name(self) -> str:
        return self._auth_secret_name

    @property
    def auth_secret_username_field(self) -> str:
        return self._auth_secret_username_field

    @property
    def auth_secret_password_field(self) -> str:
        return self._auth_secret_password_field

    @property
    def tls_enabled(self) -> bool:
        return self._tls_enabled

    @property
    def tls_secret_name(self) -> str:
        return self._tls_secret_name

    @property
    def tls_secret_cert_field(self) -> str:
        return self._tls_secret_cert_field

    ## Interface ###############################################################

    _DICT_FIELDS = [
        "hostname",
        "port",
        "auth_secret_name",
        "auth_secret_username_field",
        "auth_secret_password_field",
        "tls_secret_name",
        "tls_secret_cert_field",
    ]

    _PROVIDED_DICT_FIELDS = [
        "uri_secret",
        "uri_secret_hostname_field",
        "uri_secret_port_field",
    ]

    def to_dict(self) -> dict:
        """Return the dict representation of the object for the CR"""
        return {field: getattr(self, f"_{field}") for field in self._DICT_FIELDS}

    @classmethod
    def from_dict(cls, session: Session, config_dict: dict) -> "PostgresConnection":
        kwargs = {"session": session}
        config_dict = common.camelcase_to_snake_case(config_dict)
        uri_secret = config_dict.get("uri_secret", {})
        uri_hostname_field = config_dict.get("uri_secret_hostname_field", {})
        uri_port_field = config_dict.get("uri_secret_port_field", {})

        # First pull provided hostname/port secret if available and fill in
        # hostname/port fields into config_dict
        if uri_secret and uri_hostname_field and uri_port_field:
            # If we have provided host/port credentials, we need to extract them
            # and place these values in our config dict
            success, secret_content = session.get_object_current_state(
                "Secret", uri_secret
            )
            assert_cluster(success, f"Fetching connection secret [{uri_secret}] failed")
            assert_precondition(
                secret_content,
                f"Missing expected Secret/{uri_secret} holding [hostname] and [port]",
            )
            assert "data" in secret_content, "Got a secret without 'data'?"
            secret_content = secret_content.get("data", {})
            hostname_raw = secret_content.get(uri_hostname_field)
            port_raw = secret_content.get(uri_port_field)
            assert_config(
                None not in [hostname_raw, port_raw],
                f"Failed to find hostname/port in uri secret [{uri_secret}]",
            )
            hostname = common.b64_secret_decode(hostname_raw)
            port = common.b64_secret_decode(port_raw)

            config_dict["hostname"], config_dict["port"] = hostname, int(port)

        for field in cls._DICT_FIELDS:
            if field not in config_dict:
                raise ValueError(f"Missing required connection element [{field}]")

            # Set the kwargs (using None in place of empty strings)
            kwargs[field] = config_dict[field] or None
        return cls(**kwargs)

    ## Client Utilities ########################################################

    def get_ssl_mode(self) -> str:
        """Get Postgres SSL mode to operate in

        Returns:
            ssl_mode: str
                "require" (tls enabled) or "disable" (tls disabled)
        """
        return "require" if self.tls_enabled else "disable"

    def get_auth_username_password(self) -> Tuple[str, str]:
        """Get the current username/password pair from the auth secret if
        available

        Returns:
            username: str or None
                The plain-text username for the instance or None if not
                available
            password: str or None
                The plain-text password for the instance or None if not
                available
        """
        # If not already known, fetch from the cluster
        if None in [self._auth_username, self._auth_password]:
            secret_content = self._fetch_secret_data(self._auth_secret_name) or {}
            username = secret_content.get(self._auth_secret_username_field)
            password = secret_content.get(self._auth_secret_password_field)
            if None in [username, password]:
                log.debug2(
                    "Failed to find username/password in auth secret [%s]",
                    self._auth_secret_name,
                )
                return None, None
            self._auth_username = username
            self._auth_password = password
        return self._auth_username, self._auth_password

    def get_tls_secret_volume_mounts(
        self,
        mount_path: str = DEFAULT_TLS_VOLUME_MOUNT_PATH,
        volume_name: str = DEFAULT_TLS_VOLUME_NAME,
    ) -> List[dict]:
        """Get the list of volumeMounts entries needed to support TLS for a
        client. If TLS is not enabled, this will be an empty list.

        Args:
            mount_path: str
                A path where the tls entries should be mounted
            volume_name: str
                The name of the volume within the pod spec

        Returns:
            volume_mounts: List[dict]
                A list of dict entries for the volume mounts which can be used
                to extend other volume lists
        """
        if self._tls_enabled:
            return [dict(name=volume_name, mountPath=mount_path)]
        return []

    def get_tls_secret_volumes(
        self,
        cert_mount_path: Optional[str] = None,
        volume_name: str = DEFAULT_TLS_VOLUME_NAME,
    ) -> List[dict]:
        """Get the list of dict entries needed to support TLS for a
        client. If TLS is not enabled, this will be an empty list.

        Args:
            cert_mount_path: Optional[str]
                The name of the file that the ca cert should be mounted to
            volume_name: str
                The name of the volume within the pod spec

        Returns:
            volumes: List[dict]
                A list of dict Volume entries which can be used to extend other
                volume lists
        """
        if self._tls_enabled:
            cert_mount_path = cert_mount_path or self._tls_secret_cert_field
            return [
                dict(
                    name=volume_name,
                    secret=dict(
                        defaultMode=common.mount_mode(440),
                        secretName=self._tls_secret_name,
                        items=[
                            dict(key=self._tls_secret_cert_field, path=cert_mount_path)
                        ],
                    ),
                )
            ]
        return []

    def get_tls_cert(self) -> Optional[str]:
        """Get the current TLS certificate for a client connection if TLS is
        enabled

        If TLS is enabled, but the cert is not found, this function will raise
        an AssertionError

        Returns:
            tls_cert: str or None
                PEM encoded cert string (not base64-encoded) if found, otherwise
                None
        """
        if self._tls_enabled:
            if self._tls_cert is None:
                secret_data = self._fetch_secret_data(self._tls_secret_name) or {}
                self._tls_cert = secret_data.get(self._tls_secret_cert_field)
                assert_precondition(
                    self._tls_cert is not None, "Failed to find TLS cert"
                )
            return self._tls_cert

        return None



"""The WatchThread Class is responsible for monitoring the cluster for
resource events
"""
# Standard
from threading import Lock
from typing import Dict, List, Optional, Set
import copy
import dataclasses
import os

# Third Party
from kubernetes import watch

# First Party
import alog

# Local
from .... import config
from ....deploy_manager import DeployManagerBase, KubeEventType, KubeWatchEvent
from ....managed_object import ManagedObject
from ..filters import FilterManager, get_configured_filter
from ..leader_election import LeadershipManagerBase
from ..utils import (
    ReconcileRequest,
    ReconcileRequestType,
    ResourceId,
    WatchedResource,
    WatchRequest,
    parse_time_delta,
)
from .base import ThreadBase

log = alog.use_channel("WTCHTHRD")

# Forward declaration of ReconcileThread
RECONCILE_THREAD_TYPE = "ReconcileThread"


class WatchThread(ThreadBase):  # pylint: disable=too-many-instance-attributes
    """The WatchThread monitors the cluster for changes to a specific GroupVersionKind either
    cluster-wide or for a particular namespace. When it detects a change it checks the event
    against the registered Filters and submits a ReconcileRequest if it passes. Every resource
    that has at least one watch request gets a corresponding WatchedResource object whose main
    job is to store the current Filter status
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        reconcile_thread: RECONCILE_THREAD_TYPE,
        kind: str,
        api_version: str,
        namespace: Optional[str] = None,
        deploy_manager: DeployManagerBase = None,
        leadership_manager: LeadershipManagerBase = None,
    ):
        """Initialize a WatchThread by assigning instance variables and creating maps

        Args:
            reconcile_thread: ReconcileThread
                The reconcile thread to submit requests to
            kind: str
                The kind to watch
            api_version: str
                The api_version to watch
            namespace: Optional[str] = None
                The namespace to watch. If none then cluster-wide
            deploy_manager: DeployManagerBase = None
                The deploy_manager to watch events
            leadership_manager: LeadershipManagerBase = None
                The leadership manager to use for elections
        """
        # Setup initial variables
        self.reconcile_thread = reconcile_thread
        self.kind = kind
        self.api_version = api_version
        self.namespace = namespace

        name = f"watch_thread_{self.api_version}_{self.kind}"
        if self.namespace:
            name = name + f"_{self.namespace}"
        super().__init__(
            name=name,
            daemon=True,
            deploy_manager=deploy_manager,
            leadership_manager=leadership_manager,
        )

        # Setup kubernetes watch resource
        self.kubernetes_watch = watch.Watch()

        # Setup watched resources and request mappings. watched_resources
        # is used to track the current status of a resource in a cluster and also includes
        # the current filters. watch_request tracks all of the Controllers that have watched
        # a specific resource or groupings of resources
        self.watched_resources: Dict[str, WatchedResource] = {}
        self.watch_requests: Dict[str, Set[WatchRequest]] = {}

        # Lock for adding/gathering watch requests
        self.watch_request_lock = Lock()

        # Variables for tracking retries
        self.attempts_left = config.python_watch_manager.watch_retry_count
        self.retry_delay = parse_time_delta(
            config.python_watch_manager.watch_retry_delay or ""
        )

    def run(self):
        """The WatchThread's control loop continuously watches the DeployManager for any new
        events. For every event it gets it gathers all the WatchRequests whose `watched` value
        applies. The thread then initializes a WatchedObject if one doesn't already exist and
        tests the event against each request's Filter. Finally, it submits a ReconcileRequest
        for all events that pass
        """

        # Check for leadership and shutdown at the start
        list_resource_version = 0
        while True:
            try:
                if not self.check_preconditions():
                    log.debug("Checking preconditions failed. Shutting down")
                    return

                for event in self.deploy_manager.watch_objects(
                    self.kind,
                    self.api_version,
                    namespace=self.namespace,
                    resource_version=list_resource_version,
                    watch_manager=self.kubernetes_watch,
                ):
                    # Validate leadership on each event
                    if not self.check_preconditions():
                        log.debug("Checking preconditions failed. Shutting down")
                        return

                    resource = event.resource

                    # Gather all the watch requests which apply to this event
                    watch_requests = self._gather_resource_requests(resource)
                    if not watch_requests:
                        log.debug2("Skipping resource without requested watch")
                        self._clean_event(event)
                        continue

                    # Ensure a watched object exists for every resource
                    if resource.uid not in self.watched_resources:
                        self._create_watched_resource(resource, watch_requests)

                    # Check both global and watch specific filters
                    watch_requests = self._check_filters(
                        watch_requests, resource, event.type
                    )
                    if not watch_requests:
                        log.debug2(
                            "Skipping event %s as all requests failed filters", event
                        )
                        self._clean_event(event)
                        continue

                    # Push a reconcile request for each watch requested
                    for watch_request in watch_requests:
                        log.debug(
                            "Requesting reconcile for %s",
                            resource,
                            extra={"resource": watch_request.requester.get_resource()},
                        )
                        self._request_reconcile(event, watch_request)

                    # Clean up any resources used for the event
                    self._clean_event(event)

                # Update the resource version to only get new events
                list_resource_version = self.kubernetes_watch.resource_version
            except Exception as exc:
                log.info(
                    "Exception raised when attempting to watch %s",
                    repr(exc),
                    exc_info=exc,
                )
                if self.attempts_left <= 0:
                    log.error(
                        "Unable to start watch within %d attempts",
                        config.python_watch_manager.watch_retry_count,
                    )
                    os._exit(1)

                if not self.wait_on_precondition(self.retry_delay.total_seconds()):
                    log.debug(
                        "Checking preconditions failed during retry. Shutting down"
                    )
                    return
                self.attempts_left = self.attempts_left - 1
                log.info("Restarting watch with %d attempts left", self.attempts_left)

    ## Class Interface ###################################################

    def stop_thread(self):
        """Override stop_thread to stop the kubernetes client's Watch as well"""
        super().stop_thread()
        self.kubernetes_watch.stop()

    ## Public Interface ###################################################

    def request_watch(self, watch_request: WatchRequest):
        """Add a watch request if it doesn't exist

        Args:
            watch_request: WatchRequest
                The watch_request to add
        """
        requester_id = watch_request.requester

        # Acquire the watch request lock before starting work
        with self.watch_request_lock:
            if watch_request in self.watch_requests.get(requester_id.global_id, []):
                log.debug3("Request already added")
                return

            # Create watch request for this kind/api_version. Use global id
            # as watch thread is already namespaced/global
            log.debug3("Adding action with key %s", requester_id.global_id)
            self.watch_requests.setdefault(requester_id.global_id, set()).add(
                watch_request
            )

    ## WatchRequest Functions  ###################################################

    def _gather_resource_requests(self, resource: ManagedObject) -> List[WatchRequest]:
        """Gather the list of actions that apply to this specific Kube event based on
        the ownerRefs and the resource itself.

        Args:
            resource: ManagedObject
                The resource for this event

        Returns:
            request_list: List[WatchRequest]
                The list of watch requests that apply
        """

        request_list = []

        # Acquire the watch request lock
        with self.watch_request_lock:
            # Check if the event resource can be reconciled directly like in the case of
            # Controllers
            resource_id = ResourceId.from_resource(resource)
            for request in self.watch_requests.get(resource_id.global_id, []):
                # Check if request has a specific name and if this event matches
                if request.requester.name and request.requester.name != resource.name:
                    continue

                unique_request = copy.deepcopy(request)
                if not unique_request.requester.name:
                    unique_request.requester = dataclasses.replace(
                        unique_request.requester, name=resource_id.name
                    )

                log.debug3(
                    "Gathering request for controller %s from %s",
                    unique_request.controller_type,
                    resource_id.global_id,
                )
                request_list.append(unique_request)

            # Check for any owners watching this resource
            for owner_ref in resource.metadata.get("ownerReferences", []):
                owner_id = ResourceId.from_owner_ref(
                    owner_ref, namespace=resource_id.namespace
                )

                if owner_id.global_id not in self.watch_requests:
                    log.debug3("Skipping event with owner_key: %s", owner_id.global_id)
                    continue

                for request in self.watch_requests.get(owner_id.global_id, []):
                    # If request has a specific name then ensure it matches
                    if (
                        request.requester.name
                        and request.requester.name != owner_ref.get("name")
                    ):
                        continue

                    # If request doesn't already have a name then force
                    # this resource. This allows multiple controllers with
                    # the same kind/api_version to own the same resource
                    unique_request = copy.deepcopy(request)
                    if not unique_request.requester.name:
                        unique_request.requester = dataclasses.replace(
                            unique_request.requester, name=owner_id.name
                        )

                    log.debug3(
                        "Gathering request for controller %s from %s",
                        unique_request.controller_type,
                        owner_ref,
                    )
                    request_list.append(unique_request)

        return request_list

    def _request_reconcile(self, event: KubeWatchEvent, request: WatchRequest):
        """Request a reconcile for a kube event

        Args:
            event: KubeWatchEvent
                The KubeWatchEvent that triggered the reconcile
            request: WatchRequest
                The object that's requested a reconcile
        """

        resource = event.resource
        event_type = event.type
        requester_id = request.requester

        # If the watch request is for a different object (e.g dependent watch) then
        # fetch the correct resource to reconcile
        if (
            requester_id.kind != event.resource.kind
            or requester_id.api_version != event.resource.api_version
            or (requester_id.name and requester_id.name != event.resource.name)
        ):
            success, obj = self.deploy_manager.get_object_current_state(
                kind=requester_id.kind,
                name=requester_id.name,
                namespace=event.resource.namespace,
                api_version=requester_id.api_version,
            )
            if not success or not obj:
                log.warning(
                    "Unable to fetch owner resource %s", requester_id.get_named_id()
                )
                return

            resource = ManagedObject(obj)
            event_type = ReconcileRequestType.DEPENDENT

        # Generate the request and push one for each watched action to the reconcile thread
        request = ReconcileRequest(request.controller_type, event_type, resource)
        self.reconcile_thread.push_request(request)

    ## Watched Resource Functions  ###################################################

    def _create_watched_resource(
        self,
        resource: ManagedObject,
        watch_requests: List[WatchRequest],
    ):
        """Create a WatchedResource and initialize it's filters

        Args:
            resource: ManagedObject
                The resource being watched
            watch_requests: List[WatchRequest]
                The list of requests that apply to this resource

        """
        # update the watched resources dict
        if resource.uid in self.watched_resources:
            return

        # Setup filter dict with global filters
        filter_dict = {None: FilterManager(get_configured_filter(), resource)}
        for request in watch_requests:
            filter_dict[request.requester.get_named_id()] = FilterManager(
                request.filters, resource
            )

        # Add watched resource to mapping
        self.watched_resources[resource.uid] = WatchedResource(
            gvk=ResourceId.from_resource(resource), filters=filter_dict
        )

    def _clean_event(self, event: KubeWatchEvent):
        """Call this function after processing every event to clean any leftover resources

        Args:
            event: KubeWatchEvent
                The kube event to clean up
        """
        if event.type == KubeEventType.DELETED:
            self.watched_resources.pop(event.resource.uid, None)

    ## Filter Functions  ###################################################

    def _check_filters(
        self,
        watch_requests: List[WatchRequest],
        resource: ManagedObject,
        event: KubeEventType,
    ) -> List[WatchRequest]:
        """Check a resource and event against both global and request specific filters

        Args:
            watch_requests: List[WatchRequest]
                List of watch requests whose filters should be checked
            resource: ManagedObject
                The resource being filtered
            event: KubeEventType
                THe event type being filtered

        Returns:
            successful_requests: List[WatchRequest]
                The list of requests that passed the filter

        """

        if resource.uid not in self.watched_resources:
            return []

        # If the default watched resource filter fails then no need to
        # check any watch requests
        watched_resource = self.watched_resources[resource.uid]
        if not watched_resource.filters[None].update_and_test(resource, event):
            return []

        output_requests = []

        # Check the watch requests for any of their filters
        for request in watch_requests:
            requester_id = request.requester.get_named_id()

            # If this is the first time this watched resource has seen this request then
            # initialize the filters
            if requester_id not in watched_resource.filters:
                watched_resource.filters[requester_id] = FilterManager(
                    request.filters, resource
                )

            if not watched_resource.filters[requester_id].update_and_test(
                resource, event
            ):
                continue

            output_requests.append(request)

        return output_requests


# Keep track of watch_threads globally, This is required to keep an accurate track of
# existing watches
watch_threads: Dict[str, WatchThread] = {}


def create_resource_watch(
    watch_request: WatchRequest,
    reconcile_thread: RECONCILE_THREAD_TYPE,
    deploy_manager: DeployManagerBase,
    leadership_manager: LeadershipManagerBase,
) -> WatchThread:
    """Create or request a watch for a resource. This function will either append the request to
    an existing thread or create a new one. This function will also start the thread if any
    other watch threads have already been started.

    Args:
        watch_request: WatchRequest
            The watch request to submit
        reconcile_thread: ReconcileThread
            The ReconcileThread to submit ReconcileRequests to
        deploy_manager: DeployManagerBase
            The DeployManager to use with the Thread
        leadership_manager: LeadershipManagerBase
            The LeadershipManager to use for election

    Returns:
        watch_thread: WatchThread
            The watch_thread that is watching the request
    """
    watch_thread = None
    watched_id = watch_request.watched

    # First check for a global watch before checking for a specific namespace watch
    if watched_id.global_id in watch_threads:
        log.debug2("Found existing global watch thread for %s", watch_request)
        watch_thread = watch_threads[watched_id.global_id]

    elif watched_id.namespace and watched_id.namespaced_id in watch_threads:
        log.debug2("Found existing namespaced watch thread for %s", watch_request)
        watch_thread = watch_threads[watched_id.namespaced_id]

    # Create a watch thread if it doesn't exist
    if not watch_thread:
        log.debug2("Creating new WatchThread for %s", watch_request)
        watch_thread = WatchThread(
            reconcile_thread,
            watched_id.kind,
            watched_id.api_version,
            watched_id.namespace,
            deploy_manager,
            leadership_manager,
        )

        watch_key = watched_id.get_id()
        watch_threads[watch_key] = watch_thread

        # Only start the watch thread if another is already watching
        for thread in watch_threads.values():
            if thread.is_alive():
                watch_thread.start_thread()
                break

    # Add action to controller
    watch_thread.request_watch(watch_request)
    return watch_thread


def get_resource_watches() -> List[WatchThread]:
    """Get the list of all watch_threads

    Returns:
        list_of_watches: List[WatchThread]
            List of watch threads
    """
    return watch_threads.values()



"""
Test the functions for adding the dependency hash annotation
"""

# Standard
import hashlib
import shlex
import subprocess
import sys

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.test_helpers.helpers import (
    MockComponent,
    MockDeployManager,
    configure_logging,
    setup_session,
)
from oper8.x.utils import deps_annotation

## Helpers #####################################################################

log = alog.use_channel("TEST")
configure_logging()


def make_data_resource(
    resource_type,
    api_version="v1",
    data=None,
    metadata=None,
    name="test",
    namespace=None,
):
    """Small helper utility for creading dummy cdk8s secrets"""
    metadata = metadata or {}
    if namespace:
        metadata["namespace"] = namespace
    metadata["name"] = name
    kwargs = {"data": data or {}, "metadata": metadata}
    return dict(
        kind=resource_type,
        apiVersion=api_version,
        **kwargs,
    )


def make_secret(*args, **kwargs):
    """Small helper utility for creading dummy cdk8s Secrets"""
    return make_data_resource("Secret", *args, **kwargs)


def make_cm(*args, **kwargs):
    """Small helper utility for creading dummy cdk8s ConfigMaps"""
    return make_data_resource("ConfigMap", *args, **kwargs)


def make_pod_spec(secret_envs=None, cm_envs=None, secret_vols=None, cm_vols=None):
    secret_envs = secret_envs or {}
    cm_envs = cm_envs or {}
    secret_vols = secret_vols or {}
    cm_vols = cm_vols or {}
    return {
        "spec": {
            "containers": [
                {
                    "name": "foo",
                    "image": "foo:latest",
                    "env": [
                        {
                            "name": env_name,
                            "valueFrom": {
                                "secretKeyRef": {"name": secret_name, "key": key}
                            },
                        }
                        for env_name, (secret_name, key) in secret_envs.items()
                    ]
                    + [
                        {
                            "name": env_name,
                            "valueFrom": {
                                "configMapKeyRef": {"name": cm_name, "key": key}
                            },
                        }
                        for env_name, (cm_name, key) in cm_envs.items()
                    ],
                }
            ],
            "volumes": [
                {"name": vol_name, "secret": {"secretName": secret_name}}
                for vol_name, secret_name in secret_vols.items()
            ]
            + [
                {"name": vol_name, "configMap": {"name": cm_name}}
                for vol_name, cm_name in cm_vols.items()
            ],
        }
    }


def make_pod_empty_value_from_spec():
    return {
        "spec": {
            "containers": [
                {
                    "name": "foo",
                    "image": "foo:latest",
                    "env": [
                        {
                            "name": "FOO",
                            "valueFrom": None,
                        }
                    ],
                }
            ]
        }
    }


def make_pod(name="test-pod", namespace=None, *args, **kwargs):
    metadata = dict(name=name)
    if namespace:
        metadata["namespace"] = namespace
    return dict(
        kind="Pod",
        apiVersion="v1",
        metadata=metadata,
        **make_pod_spec(*args, **kwargs),
    )


def make_deployment(pod, name="test-deployment"):
    return dict(
        kind="Deployment",
        apiVersion="apps/v1",
        metadata=dict(name=name),
        spec=dict(selector={}, template=pod),
    )


def make_sts(pod, name="test-sts"):
    return dict(
        kind="StatefulSet",
        apiVersion="v1",
        metadata=dict(name=name),
        spec=dict(selector={}, serviceName="foobar", template=pod),
    )


def make_rs(pod, name="test-rs", chart=None, app=None):
    return dict(
        kind="ReplicaSet",
        apiVersion="apps/v1",
        metadata=dict(name=name),
        spec=dict(selector={}, template=pod),
    )


## Tests #######################################################################


#########################
## add_deps_annotation ##
#########################


def test_add_deps_annotation_pod_secret_env():
    """Test that a component with a single pod that relies on a secret in an env
    var has the annotation added to it
    """
    secret = make_secret(name="secret", data={"key": "val"})
    pod = make_pod(secret_envs={"FOO": ("secret", "key")})
    session = setup_session()

    comp = MockComponent(api_objects=[secret, pod], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, pod)
    assert (
        pod["metadata"]["annotations"].get(deps_annotation.DEPS_ANNOTATION) is not None
    )


def test_add_deps_annotation_pod_secret_vol():
    """Test that a component with a single pod that relies on a secret in a
    volume has the annotation added to it
    """
    secret = make_secret(name="secret", data={"key": "val"})
    pod = make_pod(secret_vols={"foobar": "secret"})
    session = setup_session()
    comp = MockComponent(api_objects=[secret, pod], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, pod)
    assert (
        pod["metadata"]["annotations"].get(deps_annotation.DEPS_ANNOTATION) is not None
    )


def test_add_deps_annotation_pod_cm_env():
    """Test that a component with a single pod that relies on a cm in an env var
    has the annotation added to it
    """
    cm = make_cm(name="cm", data={"key": "val"})
    pod = make_pod(cm_envs={"FOO": ("cm", "key")})
    session = setup_session()
    comp = MockComponent(api_objects=[cm, pod], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, pod)
    assert (
        pod["metadata"]["annotations"].get(deps_annotation.DEPS_ANNOTATION) is not None
    )


def test_add_deps_annotation_pod_cm_vol():
    """Test that a component with a single pod that relies on a cm in a volume
    has the annotation added to it
    """
    cm = make_cm(name="cm", data={"key": "val"})
    pod = make_pod(cm_vols={"foobar": "cm"})
    session = setup_session()
    comp = MockComponent(api_objects=[cm, pod], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, pod)
    assert (
        pod["metadata"]["annotations"].get(deps_annotation.DEPS_ANNOTATION) is not None
    )


def test_add_deps_annotation_deployment_secret():
    """Test that a component with a single deployment that relies on a secret
    has the annotation added to it
    """
    secret = make_secret(name="secret", data={"key": "val"})
    deployment = make_deployment(
        pod=make_pod_spec(secret_envs={"FOO": ("secret", "key")})
    )
    session = setup_session()
    comp = MockComponent(api_objects=[secret, deployment], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, deployment)
    assert (
        deployment["spec"]["template"]["metadata"]["annotations"].get(
            deps_annotation.DEPS_ANNOTATION
        )
        is not None
    )


def test_add_deps_annotation_statefulset_secret():
    """Test that a component with a single statefulset that relies on a secret
    has the annotation added to it
    """
    secret = make_secret(name="secret", data={"key": "val"})
    sts = make_sts(pod=make_pod_spec(secret_envs={"FOO": ("secret", "key")}))
    session = setup_session()
    comp = MockComponent(api_objects=[secret, sts], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, sts)
    assert (
        sts["spec"]["template"]["metadata"]["annotations"].get(
            deps_annotation.DEPS_ANNOTATION
        )
        is not None
    )


def test_add_deps_annotation_replicaset_secret():
    """Test that a component with a single replicaset that relies on a secret
    has the annotation added to it
    """
    secret = make_secret(name="secret", data={"key": "val"})
    rs = make_rs(pod=make_pod_spec(secret_envs={"FOO": ("secret", "key")}))
    session = setup_session()
    comp = MockComponent(api_objects=[secret, rs], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, rs)
    assert (
        rs["spec"]["template"]["metadata"]["annotations"].get(
            deps_annotation.DEPS_ANNOTATION
        )
        is not None
    )


def test_add_deps_annotation_no_pods():
    """Test that a component with no pods does not have the annotation added to
    it
    """
    secret = make_secret(name="secret", data={"key": "val"})
    session = setup_session()
    comp = MockComponent(api_objects=[secret], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, secret)
    assert (
        secret["metadata"].get("annotations", {}).get(deps_annotation.DEPS_ANNOTATION)
        is None
    )


def test_add_deps_annotation_multi_deps():
    """Test that a component with a pod that relies on multiple data resources
    (secret and configmap) has the annotation and is different than a pod which
    relies on only one of them
    """
    secret = make_secret(name="secret", data={"key": "val"})
    cm = make_secret(name="cm", data={"key": "val"})
    pod = make_pod(secret_vols={"foo": "secret"}, cm_vols={"bar": "cm"})
    session = setup_session()
    comp = MockComponent(api_objects=[secret, cm, pod], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, pod)
    assert (
        pod["metadata"]["annotations"].get(deps_annotation.DEPS_ANNOTATION) is not None
    )


def test_add_deps_annotation_external_dep():
    """Test that a component which depends on an external dependency has the
    annotation added
    """
    pod = make_pod(secret_envs={"FOO": ("secret", "key")})
    session = setup_session()
    comp = MockComponent(api_objects=[pod], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, pod)
    assert (
        pod["metadata"]["annotations"].get(deps_annotation.DEPS_ANNOTATION) is not None
    )


def test_add_deps_annotation_empty_value_from():
    """Test that an empty valueFrom will result in no dependencies"""
    deployment = make_deployment(pod=make_pod_empty_value_from_spec())
    session = setup_session()
    comp = MockComponent(api_objects=[deployment], session=session)
    comp.render_chart(session)
    deps_annotation.add_deps_annotation(comp, session, deployment)
    assert deployment["spec"]["template"].get("metadata") is None


#########################
## get_deps_annotation ##
#########################


def test_get_deps_annotation_multi_process():
    """Test that hash values are consistent across different running instances
    of the python interpreter
    """
    py_script = "; ".join(
        [
            line.strip()
            for line in """
    from watson_assistant_application.utils import deps_annotation
    import alog
    alog.configure("debug3")
    print(deps_annotation.get_deps_annotation(None, [{"key": "val"}]))
    """.split(
                "\n"
            )
            if line
        ]
    )
    cmd = f"{sys.executable} -c '{py_script}'"
    log.debug("Full command: %s", cmd)
    res1, logs1 = subprocess.Popen(
        shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ).communicate()
    res2, logs2 = subprocess.Popen(
        shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ).communicate()
    log.debug("LOGS1: %s", logs1)
    log.debug("LOGS2: %s", logs2)
    assert res1 == res2


def test_get_deps_annotation_dict():
    """Test that the deps annotation works as expected based on the content of a
    dict
    """

    secret1 = {"key1": "val1", "key2": "val2"}
    secret2 = {"key2": "val2", "key1": "val1"}
    secret3 = {"key1": "val1", "key2": "val2"}
    secret4 = {"key1": "different", "key2": "val2"}

    session = setup_session()
    hash1 = deps_annotation.get_deps_annotation(session, [secret1])
    hash2 = deps_annotation.get_deps_annotation(session, [secret2])
    hash3 = deps_annotation.get_deps_annotation(session, [secret3])
    hash4 = deps_annotation.get_deps_annotation(session, [secret4])
    hash5 = deps_annotation.get_deps_annotation(session, [secret1, secret4])
    hash6 = deps_annotation.get_deps_annotation(session, [secret2, secret4])

    # Different content dict order
    assert hash1 == hash2
    # Different labels
    assert hash1 == hash3
    # Different data content
    assert hash1 != hash4
    # Multiple entries
    assert hash5 == hash6


@pytest.mark.parametrize(
    ["operand_ns", "data_ns"],
    [
        ("test", "test"),
        ("operand", "data"),
    ],
)
def test_get_deps_annotation_resource_lookup(operand_ns, data_ns):
    """Test that the deps annotation works as expected based on the content of a
    looked up dependency, including across namespaces
    """
    secret1 = make_secret(name="secret1", namespace=data_ns)
    secret2 = make_secret(name="secret2", namespace=data_ns)
    dm = MockDeployManager(resources=[secret1, secret2])

    session = setup_session(deploy_manager=dm, namespace=operand_ns)
    hash1a = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret1")], namespace=data_ns
    )
    hash1b = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret1")], namespace=data_ns
    )
    hash2 = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret2")], namespace=data_ns
    )
    hash3 = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret1"), ("Secret", "secret2")], namespace=data_ns
    )
    hash4 = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret1"), ("Secret", "secret2")], namespace=data_ns
    )

    # Matching resource
    assert hash1a == hash1b
    # Different names
    assert hash1a != hash2
    # Multiple entries
    assert hash3 == hash4


def test_get_deps_annotation_resource_lookup_missing():
    """Test that when a dependency cannot be found, the hash is still computed.
    See the comment in the implementation for full details on why this is done,
    but the short version is that if deps are circular, rejecting this could
    result in an unrecoverable state.
    """

    secret1 = make_secret(name="secret1", metadata={"namespace": "test"})
    secret2 = make_secret(name="secret2", metadata={"namespace": "test"})
    dm = MockDeployManager(resources=[secret1])

    session = setup_session(deploy_manager=dm)

    # Hash before secret2 exists
    hash1 = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret1"), ("Secret", "secret2")]
    )

    # Deploy secret2 and make sure the hash changes
    dm.deploy([secret2])
    hash2 = deps_annotation.get_deps_annotation(
        session, [("Secret", "secret1"), ("Secret", "secret2")]
    )
    assert hash1 != hash2


def test_get_deps_annotation_resource_lookup_fail():
    """Test that if a dependency lookup fails, an error is thrown"""
    dm = MockDeployManager(get_state_fail=True, auto_enable=False)
    session = setup_session(deploy_manager=dm)
    dm.enable_mocks()
    with pytest.raises(AssertionError):
        deps_annotation.get_deps_annotation(session, [("Secret", "secret1")])


def test_get_deps_annotation_object_types():
    """Test that the deps annotation works as expected for various data types"""

    # Secrets
    secret1 = make_secret(
        name="secret",
        metadata={"labels": {"foo": "bar"}},
        data={"key1": "val1", "key2": "val2"},
    )

    # Configmaps
    cm1 = make_cm(
        name="cm",
        metadata={"labels": {"foo": "bar"}},
        data={"key1": "val1", "key2": "val2"},
    )
    cm2 = make_cm(
        name="cm2",
        metadata={"labels": {"foo": "bar"}},
        data={"key1": "val1", "key2": "val2"},
    )
    cm3 = make_cm(
        name="cm",
        metadata={"labels": {"foo": "bar"}},
        data={"key1": "val1", "key2": "changed"},
    )
    cm4 = make_cm(
        name="cm",
        metadata={"labels": {"foo": "bar4"}},
        data={"key1": "val1", "key2": "val2"},
    )

    # Other
    foo1 = {
        "kind": "Foo",
        "metadata": {"name": "foo", "labels": {"foo": "bar"}},
        "key": "val",
    }
    foo2 = {
        "kind": "Foo",
        "metadata": {"name": "bar", "labels": {"foo": "bar"}},
        "key": "val",
    }
    foo3 = {
        "kind": "Foo",
        "metadata": {"name": "foo", "labels": {"foo": "bar"}},
        "key": "different",
    }
    foo4 = {
        "kind": "Foo",
        "metadata": {"name": "bar", "labels": {"foo": "bar4"}},
        "key": "val",
    }

    session = setup_session()

    with alog.ContextLog(log.debug, "Kind different"):
        assert deps_annotation.get_deps_annotation(
            session, [secret1]
        ) != deps_annotation.get_deps_annotation(session, [cm1])

    with alog.ContextLog(log.debug, "ConfigMap"):
        with alog.ContextLog(log.debug, "Match"):
            assert deps_annotation.get_deps_annotation(
                session, [cm1]
            ) == deps_annotation.get_deps_annotation(session, [cm1])
        with alog.ContextLog(log.debug, "Name different"):
            assert deps_annotation.get_deps_annotation(
                session, [cm1]
            ) != deps_annotation.get_deps_annotation(session, [cm2])
        with alog.ContextLog(log.debug, "Val different"):
            assert deps_annotation.get_deps_annotation(
                session, [cm1]
            ) != deps_annotation.get_deps_annotation(session, [cm3])
        with alog.ContextLog(log.debug, "Data Match w/ label different"):
            assert deps_annotation.get_deps_annotation(
                session, [cm1]
            ) == deps_annotation.get_deps_annotation(session, [cm4])

    with alog.ContextLog(log.debug, "Foo"):
        with alog.ContextLog(log.debug, "Match"):
            assert deps_annotation.get_deps_annotation(
                session, [foo1]
            ) == deps_annotation.get_deps_annotation(session, [foo1])
        with alog.ContextLog(log.debug, "Name different"):
            assert deps_annotation.get_deps_annotation(
                session, [foo1]
            ) != deps_annotation.get_deps_annotation(session, [foo2])
        with alog.ContextLog(log.debug, "Val different"):
            assert deps_annotation.get_deps_annotation(
                session, [foo1]
            ) != deps_annotation.get_deps_annotation(session, [foo3])
        with alog.ContextLog(log.debug, "Label different"):
            assert deps_annotation.get_deps_annotation(
                session, [foo1]
            ) != deps_annotation.get_deps_annotation(session, [foo4])
